# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from pathlib import Path
import subprocess
import platform
import sys
import logging
import os
import stat
import uuid
from kqcircuits.defaults import TMP_PATH, SCRIPTS_PATH, KQC_REMOTE_TMP_PATH

logging.basicConfig(level=logging.WARN, stream=sys.stdout)

def _prepare_remote_tmp(ssh_login, kqc_remote_tmp_path):
    """
    Internal helper function to create remote tmp directory if it doesnt exist and raise error if its not empty

    Args:
        ssh_login              (str): ssh login info "user@hostname"
        kqc_remote_tmp_path    (str): current run tmp directory on remote
    """
    ssh_cmd = f"ssh {ssh_login} 'mkdir -p {kqc_remote_tmp_path} && ! {{ ls -1qA {kqc_remote_tmp_path} | grep -q . ; }}'"
    is_empty = subprocess.call(ssh_cmd, shell=True)
    if is_empty:
        logging.error(f"Your remote tmp folder {kqc_remote_tmp_path} is not empty!")
        logging.error("Either delete its contents manually or use another directory")
        sys.exit()

def _get_sbatch_time(export_tmp_paths) -> int:
    """
    Internal helper function to extract sbatch time limit from simulation.sh files

    Args:
        export_tmp_paths  (list[str]): list of export paths

    Returns:
        sbatch_time (int) Number of seconds for setting ssh timeout.
                          This is equal to total amount of time reserved for all batch jobs sent to remote
    """
    def _get_single_sbatch_time(simulation_script):
        with open(simulation_script , 'r') as f:
            for line in f:
                res = line.strip().partition("#SBATCH --time=")[2]
                if len(res) == 8:
                    times = res.split(':')
                    if len(times)==3:
                        return 3600*int(times[0]) + 60*int(times[1]) + int(times[2])
        return 0

    sbatch_time = 0
    for d in export_tmp_paths:
        t_meshes = _get_single_sbatch_time(Path(d)/"simulation_meshes.sh")
        t = _get_single_sbatch_time(Path(d)/"simulation.sh")
        if t == 0 or t_meshes == 0:
            logging.warning("Could not extract the sbatch time limit from simulation.sh or simulation_meshes.sh")
            logging.warning("Wait script timeout of 60 minutes will be used")
            return 3600
        else:
            sbatch_time += t

    return sbatch_time

def _remote_run(ssh_login: str,
                export_tmp_paths: list,
                kqc_remote_tmp_path: str,
                detach_simulation: bool,
                poll_interval: int):
    """
    Internal helper function to copy and run simulations to remote and back

    Args:
        ssh_login              (str): ssh login info "user@hostname"
        export_tmp_paths (list[str]): list of local tmp simulation export paths for the simulations to be run
        kqc_remote_tmp_path    (str): tmp directory on remote
        detach_simulation      (bool): Detach the remote simulation from terminal, not waiting for it to finish
        poll_interval           (int): Polling interval in seconds when waiting for the remote simulation to finish
    """
    if platform.system() == 'Windows':  # Windows
        logging.error("Connecting to remote host not supported on Windows")
        sys.exit()
    elif platform.system() == 'Darwin':  # macOS
        logging.error("Connecting to remote host not supported on Mac OS")
        sys.exit()
    # Else linux

    # set defaults
    if kqc_remote_tmp_path is None:
        kqc_remote_tmp_path = KQC_REMOTE_TMP_PATH

    if poll_interval is None:
        poll_interval = 60

    # Check if we sugin sbatch by checking if all export folders have `simulation_meshes.sh`
    if not all(((Path(d)/"simulation_meshes.sh").is_file() for d in export_tmp_paths)):
        logging.error('Simulation not exported with "sbatch" (simulation_meshes.sh does not exist)' )
        sys.exit()

    # Add uuid to the remote path for this run
    # Allows simultaneous calls to "kqc sim --remote"
    run_uuid = str(uuid.uuid4())
    kqc_remote_tmp_path = str(Path(kqc_remote_tmp_path) / ('run_' + run_uuid))
    # Create remote tmp if it doesnt exist, and check that its empty
    _prepare_remote_tmp(ssh_login,kqc_remote_tmp_path)

    dirs_remote = [str(Path(kqc_remote_tmp_path) / str(Path(d).name)) for d in export_tmp_paths]

    print('\nFEM simulations prepared successfully.\n'
          'Submitting the following simulations to the remote host (can take some time):', flush=True)
    for d1,d2 in zip(export_tmp_paths, dirs_remote):
        print(f"{d1}   --->   user@remote:{d2}", flush=True)
    print("\n", flush=True)


    remote_script_name = f"remote_simulation_{run_uuid}.sh"
    remote_simulation_script = str(TMP_PATH / remote_script_name)
    simlist = ' '.join(dirs_remote)

    remote_simulation = f"""#!/bin/bash
    sim_list=({simlist})

    for i in "${{sim_list[@]}}"; do
        cd "${{i}}" || exit
        RES=$(sbatch -J "{run_uuid}" ./simulation_meshes.sh) && sbatch -d afterok:${{RES##* }} -J "{run_uuid}" ./simulation.sh
    done;
    """
    with open(remote_simulation_script, "w") as file:
        file.write(remote_simulation)

    os.chmod(remote_simulation_script, os.stat(remote_simulation_script).st_mode | stat.S_IEXEC)

    # hardcoded option for copying back the vtu and pvtu files from remote
    copy_vtus = False
    skip_patterns = '-name "mesh.*" -o -name "*.msh" -o -name "scripts" -o -name "partitioning.*"'
    if not copy_vtus:
        skip_patterns = skip_patterns + ' -o -name "*.vtu" -o -name "*.pvtu"'
    skip_patterns = r'\( ' + skip_patterns +  r' \)'

    poll_interval_str = f"{poll_interval}s" if poll_interval <= 60 else f"{round(float(poll_interval)/60, 1)} min"
    # Write script to run in the background and copy results back once simulation is finished
    wait_and_copy_back_script = str(TMP_PATH / f"fetch_remote_simulation_data_{run_uuid}.sh")
    wait_and_copy_back = f"""#!/bin/bash
    set -e
    echo "\n---------START-WAIT-SCRIPT---------"
    echo "Simulations sent to queue at:"
    date +"%d-%m-%y %T"
    sleep 5
    echo "\nExplanation of Slurm job states"
    echo "ALL: Number of all unfinished jobs"
    echo " PD: Number of pending jobs"
    echo "  R: Number of currently running jobs\n"

    jobs_states=$(ssh {ssh_login} "squeue -h -n {run_uuid} -o%t")
    n_all=$(echo "$jobs_states" | wc -w)
    n_pd=$(echo "$jobs_states" | grep PD | wc -w)
    n_run=$(echo "$jobs_states" | grep R | wc -w)

    while [[ "$n_all" -gt 0  && $counter -le {_get_sbatch_time(export_tmp_paths)} ]]
    do
        echo -n "[ALL: $n_all, PD: $n_pd, R: $n_run] " && date +"%d-%m-%y %T"

        if [[ "$n_run" -gt 0 ]]
        then
            counter=$((counter + {poll_interval}))
        fi

        sleep {poll_interval}

        jobs_states=$(ssh {ssh_login} "squeue -h -n {run_uuid} -o%t")
        n_all=$(echo "$jobs_states" | wc -w)
        n_pd=$(echo "$jobs_states" | grep PD | wc -w)
        n_run=$(echo "$jobs_states" | grep R | wc -w)
    done

    ssh {ssh_login} 'find {kqc_remote_tmp_path} {skip_patterns} -exec rm -r "{{}}" +'

    scp -r -q {ssh_login}:"{simlist}" {str(TMP_PATH)}
    ssh {ssh_login} "rm -r {kqc_remote_tmp_path}"
    echo "\nSimulations finished at:"
    date +"%d-%m-%y %T"
    echo "---------STOP-WAIT-SCRIPT---------"
    rm -- "$0"
    """

    with open(wait_and_copy_back_script, "w") as file:
        file.write(wait_and_copy_back)
    os.chmod(wait_and_copy_back_script, os.stat(wait_and_copy_back_script).st_mode | stat.S_IEXEC)


    try:
        copy_cmd = ['scp', '-r', '-q'] + export_tmp_paths + \
                   [remote_simulation_script, ssh_login + ':' + kqc_remote_tmp_path]
        run_cmd =  f"""ssh {ssh_login} -tt -q 'bash -l -c "cd {kqc_remote_tmp_path} && ./{remote_script_name}"'"""
        # COPY (dirs_local) -> (dirs_remote)
        subprocess.check_call(copy_cmd)
        # Remove remote simulation script from local tmp folder
        subprocess.check_call(['rm',  remote_simulation_script])
        # Force to use login shell on remote to get correct env variables
        subprocess.check_call(run_cmd, shell=True)

        print(f"Simulations started and connection to remote closed.\n"
              f"Starting a script to follow the submitted jobs with {poll_interval_str} interval", flush=True)
        # start ssh poll and wait script
        if detach_simulation:
            nohup_file = str(TMP_PATH / f"nohup_{run_uuid}.out")
            wait_and_copy_back_script = "nohup " + wait_and_copy_back_script + " > " + nohup_file + " 2>&1 &"

        subprocess.check_call(wait_and_copy_back_script, shell=True)

        if detach_simulation:
            print("Simulation wait script sent to background. You can follow the job state with"
                  f" 'watch cat {nohup_file}'", flush=True)

    except Exception as exc:
        raise RuntimeError("Starting remote run failed. Please manually fetch and delete data from remote") from exc

def _allowed_simulations():
    """
    Helper to list allowed simulations, simulations scripts and tmp directory.

    Returns:
        tuple containing

            * allowed_simulations (list[str]): List of allowed simulation names
                                               cpw_fem_xsection.py added as an exception to this
            * simdir                    (str): Path to simulation scripts directory under KQCircuits
            * tmpdir                    (str): Path to tmp directory under KQCircuits
    """

    tmpdir = str(TMP_PATH)

    if 'KQCircuits' not in tmpdir:
        logging.error('Non-default tmp path. \
                      Check that the KQC_ROOT_PATH environment variable is properly set')
        sys.exit()

    simdir = str(SCRIPTS_PATH / "simulations")

    if 'KQCircuits' not in simdir:
        logging.error('Non-default simulations path. \
                      Check that the KQC_ROOT_PATH environment variable is properly set')
        sys.exit()

    allowed_simulations = ['cpw_fem_xsection.py']
    for f in os.listdir(simdir):
        if os.path.isfile(os.path.join(simdir, f)):
            allowed_simulations.append(f)

    return (allowed_simulations, simdir, tmpdir)


def remote_export_and_run(ssh_login: str,
                          kqc_remote_tmp_path: str=None,
                          detach_simulation :bool=False,
                          poll_interval: int=None,
                          args=None):
    """
    Exports locally and runs KQC simulations on a remote host. Froced to use no GUI (--quiet, -q option)

    Args:
        ssh_login              (str): ssh login info "user@hostname"
        kqc_remote_tmp_path    (str): tmp directory on remote
        detach_simulation     (bool): Detach the remote simulation from terminal, not waiting for it to finish
        poll_interval           (int): Polling interval in seconds when waiting for the remote simulation to finish
        args                  (list): a list of strings:
                                        - If starts with a letter and ends with ".py"  -> export script
                                        - If starts with "-" or "--"                   -> script option
    """

    allowed_simulations, simdir, tmpdir = _allowed_simulations()

    export_scripts = []
    # By default use --use-sbatch and -q flags. Note that the export script must support this
    args_for_script = ['--use-sbatch', '-q']
    # Separate export script filenames and script arguments
    for arg in args:
        if arg.startswith('-'):
            args_for_script.append(arg)
        else:
            if arg.endswith('.py'):
                arg_filename = Path(arg).name
                arg_path = Path(simdir) / arg_filename
                if arg_filename != arg:
                    logging.warning(f"Concatenating the path to its final component and search in {simdir} instead")
                    logging.warning(f"{arg} -> {arg_filename} -> {arg_path}")
                if arg in allowed_simulations:
                    export_scripts.append(arg_path)
                else:
                    logging.warning(f"Skipping unkown simulation: {arg}")
            else:
                logging.warning(f"Skipping unkown argument: {arg}")


    export_tmp_paths = [str(Path(tmpdir) / str(script.stem)) for script in export_scripts]

    if len(export_scripts) == 0:
        logging.error("No valid simulation script provided in remote_export_and_run")
        sys.exit()

    # Export simulation files locally
    for export_path, export_script in zip(export_tmp_paths, export_scripts):
        export_cmd = [sys.executable, export_script,
            '--simulation-export-path', str(export_path)] + args_for_script
        subprocess.call(export_cmd)

    # Run on remote
    _remote_run(ssh_login,
                export_tmp_paths,
                kqc_remote_tmp_path,
                detach_simulation,
                poll_interval)


def remote_run_only(ssh_login: str,
                    kqc_remote_tmp_path: str=None,
                    detach_simulation: bool=False,
                    poll_interval: int=None,
                    export_tmp_dirs: list=None):
    """
    Runs already locally exported simulations on remote host

    Args:
        ssh_login              (str): ssh login info "user@hostname"
        kqc_remote_tmp_path    (str): tmp directory on remote
        detach_simulation      (bool): Detach the remote simulation from terminal, not waiting for it to finish
        poll_interval           (int): Polling interval in seconds when waiting for the remote simulation to finish
        export_tmp_dirs   (list[str]): list of local tmp simulation folder names
                                       Could contain other arguments from console script which are filtered out
    """
    allowed_simulations, _, tmpdir = _allowed_simulations()

    allowed_simulations = [str(Path(sim).stem) for sim in allowed_simulations]

    paths_filtered = []
    if export_tmp_dirs is not None:
        for p in export_tmp_dirs:
            # Only allow paths directly under KQC tmp path
            p_filename = Path(p).name
            p_path = Path(tmpdir) / p_filename
            if p_filename != p:
                logging.warning(f"Concatenating the path to its final component and search in {tmpdir} instead")
                logging.warning(f"{p} -> {p_filename} -> {p_path}")
            if os.path.isdir(p_path):
                if any((sim in p  for sim in allowed_simulations)):
                    paths_filtered.append(p_path)

    if len(paths_filtered) == 0:
        logging.error("No valid simulation export paths provided in remote_run_only")
        sys.exit()

    # Run on remote
    _remote_run(ssh_login,
                paths_filtered,
                kqc_remote_tmp_path,
                detach_simulation,
                poll_interval)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).



from kqcircuits.pya_resolver import pya
from kqcircuits.chips.chip import Chip
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.qubits.swissmon import Swissmon
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare



class Simple(Chip):
    """The PCell declaration for a very simple chip.

       Contains a small number of elements connected in a simple, direct way.
    """

    name_chip = Param(pdt.TypeString, "Name of the chip", "Simple")

    def build(self):

        # Launcher
        launchers = self.produce_launchers("SMA8", enabled=["WN", "EN", "SE"])

        # Swissmon
        _pos = pya.DTrans(3, False, 2500, launchers["WN"][0].y)
        _, swissmon_refs = self.insert_cell(Swissmon, _pos, cpl_length=[0, 150, 0])

        # Two FingerCapacitorSquare instances
        cap_cell = self.add_element(FingerCapacitorSquare, finger_width=20, finger_length=100, finger_number=7)
        _, cap1_refs = self.insert_cell(cap_cell, pya.DTrans(5000.0, launchers["WN"][0].y), "C1")
        self.insert_cell(cap_cell, pya.DTrans(3, False, launchers["SE"][0].x, 5000.0), "C2")

        # WaveguideCoplanarSplitter
        _pos = pya.DTrans(launchers["SE"][0].x, launchers["WN"][0].y)
        _, tcross_refs = self.insert_cell(WaveguideCoplanarSplitter, _pos, **t_cross_parameters(
            a=self.a, b=self.b, a2=self.a, b2=self.b))

        # Waveguides: WN -> Swissmon -> FingerCapacitorSquare -> WaveguideCoplanarSplitter -> EN
        self.insert_cell(WaveguideCoplanar, path=pya.DPath([launchers["WN"][0], swissmon_refs["port_flux"]], 1))
        self.insert_cell(WaveguideCoplanar, path=pya.DPath([swissmon_refs["port_cplr1"], cap1_refs["port_a"]], 1))
        self.insert_cell(WaveguideCoplanar, path=pya.DPath([cap1_refs["port_b"], tcross_refs["port_left"]], 1))
        self.insert_cell(WaveguideCoplanar, path=pya.DPath([tcross_refs["port_right"], launchers["EN"][0]], 1))

        # Waveguides: TCross' bottom -> second capacitor -> SE
        refs = self.refpoints
        self.insert_cell(WaveguideCoplanar, path=pya.DPath([tcross_refs["port_bottom"], refs["C2_port_a"]], 1))
        self.insert_cell(WaveguideCoplanar, path=pya.DPath([refs["C2_port_b"], launchers["SE"][0]], 1))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.chip import Chip
from kqcircuits.pya_resolver import pya


class Empty(Chip):
    """Chip with almost all ground metal removed, used for EBL tests."""

    def make_empty_area(self):
        d1 = float(self.frames_dice_width[0]) + self.dice_grid_margin
        d2 = 2000

        empty_area = pya.DPolygon([
            pya.DPoint(d1, d2),
            pya.DPoint(d1, 10000 - d2),
            pya.DPoint(d2, 10000 - d2),
            pya.DPoint(d2, 10000 - d1),
            pya.DPoint(10000-d2, 10000 - d1),
            pya.DPoint(10000-d2, 10000 - d2),
            pya.DPoint(10000-d1, 10000 - d2),
            pya.DPoint(10000-d1, d2),
            pya.DPoint(10000-d2, d2),
            pya.DPoint(10000-d2, d1),
            pya.DPoint(d2, d1),
            pya.DPoint(d2, d2),
        ])

        return empty_area

    def build(self):
        empty_area = self.make_empty_area()
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(empty_area)
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(empty_area)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.chip import Chip
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare
from kqcircuits.elements.finger_capacitor_taper import FingerCapacitorTaper
from kqcircuits.elements.meander import Meander
from kqcircuits.junctions.junction import Junction
from kqcircuits.qubits.swissmon import Swissmon
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import point_shift_along_vector
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.chips.demo import Demo

@add_parameters_from(Demo, "readout_res_lengths", "include_couplers", frames_enabled=[0, 1])
@add_parameters_from(Junction, "junction_type")
class DemoTwoface(Chip):
    """Demonstration chip for 3D-integration (multi-face) features."""

    name_chip = Param(pdt.TypeString, "Name of the chip", "DT")

    def build(self):

        launcher_assignments = {
            # N
            2: "FL-QB1",
            3: "PL-1-IN",
            4: "PL-2-IN",
            5: "FL-QB2",
            # E
            7: "DL-QB2",
            12: "DL-QB4",
            # S
            14: "FL-QB4",
            15: "PL-2-OUT",
            16: "PL-1-OUT",
            17: "FL-QB3",
            # W
            19: "DL-QB3",
            24: "DL-QB1",
        }
        self.produce_launchers("ARD24", launcher_assignments)

        self.produce_qubits()
        if self.include_couplers:
            self.produce_couplers()
        self.produce_control_lines()
        self.produce_readout_structures()
        self.produce_probelines()

    def produce_qubits(self):
        dist_x = 2000  # distance from bottom chip edge
        dist_y = 3200
        self.produce_qubit(pya.DTrans(0, True, dist_x, 1e4 - dist_y), "QB1")
        self.produce_qubit(pya.DTrans(2, False, 1e4 - dist_x, 1e4 - dist_y), "QB2")
        self.produce_qubit(pya.DTrans(0, False, dist_x, dist_y), "QB3")
        self.produce_qubit(pya.DTrans(2, True, 1e4 - dist_x, dist_y), "QB4")

    def produce_qubit(self, trans, inst_name):
        self.insert_cell(Swissmon, trans, inst_name,
            cpl_length=[120, 120, 120],
            port_width=[4, 10, 4],
        )

    def produce_couplers(self):
        self.produce_coupler(1, 2)
        self.produce_coupler(4, 3)

    def produce_coupler(self, qubit_a_nr, qubit_b_nr):
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["QB{}_port_cplr2".format(qubit_a_nr)]),
            Node(self.refpoints["QB{}_port_cplr2".format(qubit_b_nr)]),
        ], a=4, b=9)

    def produce_control_lines(self):
        for qubit_nr in [1, 2, 3, 4]:
            self.produce_driveline(qubit_nr)
            self.produce_fluxline(qubit_nr)

    def produce_driveline(self, qubit_nr):
        port_drive = self.refpoints["QB{}_port_drive".format(qubit_nr)]
        port_corner = self.refpoints["DL-QB{}_port_corner".format(qubit_nr)]
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["DL-QB{}_base".format(qubit_nr)]),
            Node(port_corner),
            Node(port_drive),
        ], term2=self.b)

    def produce_fluxline(self, qubit_nr):
        port_corner = self.refpoints["FL-QB{}_port_corner".format(qubit_nr)]
        port_flux = self.refpoints["QB{}_port_flux".format(qubit_nr)]
        shift = 1500 if qubit_nr in [3, 4] else -1500
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["FL-QB{}_base".format(qubit_nr)]),
            Node(port_corner + pya.DPoint(0, shift)),
            Node((port_flux.x, port_corner.y + shift)),
            Node(self.refpoints["QB{}_port_flux".format(qubit_nr)])
        ])

    def produce_readout_structures(self):
        self.produce_readout_structure(1, False, 7)
        self.produce_readout_structure(2, True, 6)
        self.produce_readout_structure(3, False, 5)
        self.produce_readout_structure(4, True, 4)

    def produce_readout_structure(self, qubit_nr, mirrored, cap_finger_nr):

        # non-meandering part of the resonator

        point_1 = self.refpoints["QB{}_port_cplr1".format(qubit_nr)]
        point_2 = point_shift_along_vector(self.refpoints["QB{}_port_cplr1".format(qubit_nr)],
                                           self.refpoints["QB{}_base".format(qubit_nr)], -700)
        point_3 = point_2 + pya.DPoint(-400 if mirrored else 400, 0)
        point_4 = point_3 + pya.DPoint(-100 if mirrored else 100, 0)

        waveguide_inst, _ = self.insert_cell(WaveguideComposite, nodes=[
            Node(point_1),
            Node(point_2),
            Node(point_3, face_id=self.face_ids[1]),
            Node(point_4),
        ])
        length_nonmeander = waveguide_inst.cell.length()

        # meandering part of the resonator

        meander_start = point_4
        meander_end = point_4 + pya.DPoint(-1300 if mirrored else 1300, 0)

        self.insert_cell(Meander,
            start=meander_start,
            end=meander_end,
            length=float(self.readout_res_lengths[qubit_nr - 1]) - length_nonmeander,
            meanders=5,
            face_ids=[self.face_ids[1]]
        )

        # capacitor and tcross waveguide connecting resonator to probeline

        if mirrored:
            cap_rot = 2
            tcross_rot = 1
        else:
            cap_rot = 0
            tcross_rot = 3
        cap_cell = self.add_element(FingerCapacitorSquare,
            finger_number=cap_finger_nr,
            face_ids=[self.face_ids[1]]
        )
        cap_ref_rel = self.get_refpoints(cap_cell, pya.DTrans(cap_rot, False, 0, 0))
        cap_trans = pya.DTrans(cap_rot, False, meander_end + cap_ref_rel["base"] - cap_ref_rel["port_a"])
        _, cap_ref_abs = self.insert_cell(cap_cell, cap_trans)

        tcross_cell = self.add_element(WaveguideCoplanarSplitter, **t_cross_parameters(
            a=self.a, b=self.b, a2=self.a, b2=self.b, length_extra_side=30, face_ids=[self.face_ids[1]]))
        tcross_ref_rel = self.get_refpoints(tcross_cell, pya.DTrans(tcross_rot, False, 0, 0))
        tcross_trans = pya.DTrans(tcross_rot, False, cap_ref_abs["port_b"] - tcross_ref_rel["port_bottom"])
        self.insert_cell(tcross_cell, tcross_trans, inst_name="PL{}".format(qubit_nr), label_trans=pya.DCplxTrans(0.2))

    def produce_probelines(self):
        self.produce_probeline("PL-1", 1, 3, True, 4)
        self.produce_probeline("PL-2", 2, 4, False, 6)

    def produce_probeline(self, probeline_name, qubit_a_nr, qubit_b_nr, mirrored, cap_finger_nr):

        cap_cell = self.add_element(FingerCapacitorTaper,
            finger_number=cap_finger_nr,
            taper_length=20,
            face_ids=[self.face_ids[1]]
        )
        cap_trans = pya.DTrans(3, False, self.refpoints["PL{}_port_left".format(qubit_a_nr)] + pya.DPoint(0, 700))
        _, cap_ref_abs = self.insert_cell(cap_cell, cap_trans)

        # segment 1
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["{}-IN_base".format(probeline_name)]),
            Node(self.refpoints["{}-IN_port_corner".format(probeline_name)] + pya.DPoint(0, -1000)),
            Node((self.refpoints["PL{}_port_left".format(qubit_a_nr)].x,
                  self.refpoints["{}-IN_port_corner".format(probeline_name)].y - 1000)),
            Node(cap_ref_abs["port_a"] + pya.DPoint(0, 700), face_id=self.face_ids[1]),
            Node(cap_ref_abs["port_a"]),
        ])

        port_1_side = "left" if mirrored else "right"
        port_2_side = "right" if mirrored else "left"

        # segment 2
        self.insert_cell(WaveguideComposite, nodes=[
            Node(cap_ref_abs["port_b"]),
            Node(self.refpoints["PL{}_port_{}".format(qubit_a_nr, port_1_side)]),
        ], face_ids=[self.face_ids[1]])

        # segment 3
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["PL{}_port_{}".format(qubit_a_nr, port_2_side)]),
            Node(self.refpoints["PL{}_port_{}".format(qubit_b_nr, port_1_side)]),
        ], face_ids=[self.face_ids[1]])

        # segment 4
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["{}-OUT_base".format(probeline_name)]),
            Node(self.refpoints["{}-OUT_port_corner".format(probeline_name)] + pya.DPoint(0, 1000)),
            Node((self.refpoints["PL{}_port_right".format(qubit_b_nr)].x,
                  self.refpoints["{}-OUT_port_corner".format(probeline_name)].y + 1000)),
            Node(self.refpoints["PL{}_port_right".format(qubit_b_nr)] + pya.DPoint(0, -1400), face_id=self.face_ids[1]),
            Node(self.refpoints["PL{}_port_{}".format(qubit_b_nr, port_2_side)]),
        ])



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt


class AirbridgeRectangular(Airbridge):
    """PCell declaration for a rectangular airbridge.

    Origin is at the geometric center. The airbridge is in vertical direction.

    Bottom parts of pads in bottom layer, bridge and top parts of pads in top layer. Pads and bridge are rectangular.
    Refpoints "port_a" and "port_b" at top pad points closest to origin.

    .. MARKERS_FOR_PNG 0,30 0,1 -10,30 -2,14
    """

    default_type = "Airbridge Rectangular"
    pad_extra = Param(pdt.TypeDouble, "Bottom pad extra", 2, unit="μm")

    def build(self):
        # shorthand
        (w, h, l, e) = (self.bridge_width, self.pad_length, self.bridge_length, self.pad_extra)

        pts = [
            pya.DPoint(-w / 2 - e, l / 2),
            pya.DPoint(-w / 2 - e, h + l / 2),
            pya.DPoint(w / 2 + e, h + l / 2),
            pya.DPoint(w / 2 + e, l / 2),
        ]
        self._produce_bottom_pads(pts)

        # top layer
        pts = [
            pya.DPoint(-w / 2, h + l / 2 - e),
            pya.DPoint(w / 2, h + l / 2 - e),
            pya.DPoint(w / 2, -h - l / 2 + e),
            pya.DPoint(-w / 2, -h - l / 2 + e),
        ]
        self._produce_top_pads_and_bridge(pts)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import json
import logging
import os.path
from os import cpu_count

from kqcircuits.defaults import default_layers, default_netlist_breakdown, default_faces
from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import get_cell_path_length
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder

log = logging.getLogger(__name__)


def export_cell_netlist(cell, filename, pcell=None, alt_netlists=None):
    """Exports netlist(s) in JSON into file(s).

    The file will have four sections:
    ``{"nets": {...}, "subcircuits": {...}, "circuits": {...}, "chip": {...}}``

    KLayout's `terminology <https://www.klayout.de/doc-qt5/manual/lvs_overview.html>`__ differs
    from the one used in typical EDA tools where we have components (resistors, capacitors, etc.),
    pins (the endpoints of components) and nets (i.e. wires between pins). Components are PCell
    instances, a.k.a. cells, these are called subcircuits in the netlist file.

    The main conceptual difference is that waveguides, that would be analogous to wires, are also
    treated as components. Consequently, a net in the ``nets`` section usually contains exactly two
    overlapping pins that belong to two different components each identified by a unique
    ``subcircuit_id``. One of these is almost always a waveguide. Unconnected pins are not shown
    except for Launchers.

    The ``subcircuits`` section is a dictionary of the used cells: ``<subcircuit_id>: {"cell_name":
    "...", "subcircuit_location": {"_pya_type": "DPoint", "x": <x>, "y": <y>}, ...}``.
    Where ``cell_name`` is the name of the used Element optionally appended with ``$<n>`` if there
    are more than one Elements of the same type. Different instances of the same cell will have
    different ``subcircuit_id`` but identical ``cell_name``. ``subcircuit_location`` defines the
    center of the bounding box of the subcircuit's geometry in ``base_metal_gap_wo_grid`` layer,
    while ``subcircuit_origin`` defines the center of the bounding box of netlist ports of the cell.

    The ``circuits`` section maps ``cell_name`` to a dictionary of the named Element's parameters.

    If the Cell object is a Chip, the ``chip`` section contains bounding boxes of each face in the chip.

    This function may generate alternative netlists too as specified in the ``alt_netlists``
    dictionary. The keys should be tags that get added to the generated netlist filenames and the
    values are the corresponding Element breakdown lists used to generate them. The default netlist
    is generated regadless of this parameter.

    Args:
        cell: pya Cell object
        filename: absolute path as convertible to string
        pcell: pya PCell object. If None, an attempt is made to treat cell as pcell
        alt_netlists: optional dictionary of file name postfixes and element breakdown lists
    """
    if pcell is None:
        pcell = cell

    _export_cell_netlist_breakdown(cell, filename, pcell, default_netlist_breakdown)

    if isinstance(alt_netlists, dict):
        fn, ext = os.path.splitext(filename)
        for tag, breakdown in alt_netlists.items():
            fnt = f"{fn}_{tag}{ext}"
            _export_cell_netlist_breakdown(cell, fnt, pcell, breakdown)


def _export_cell_netlist_breakdown(cell, filename, pcell, breakdown_list):
    """A helper function of ``export_cell_netlist``, processes a single breakdown list."""

    # get LayoutToNetlist object
    layout = cell.layout()
    faces_with_ports = [face_id for face_id in default_faces if f"{face_id}_ports" in default_layers]
    port_layers = [layout.layer(default_layers[f"{face_id}_ports"]) for face_id in faces_with_ports]
    shapes_iter = pya.RecursiveShapeIterator(layout, cell, port_layers)
    ltn = pya.LayoutToNetlist(shapes_iter)
    # text_enlargement>0 converts the texts into boxes so that their overlaps are detected as connections
    ltn.dss().text_enlargement = 1
    # parallel processing
    ltn.threads = cpu_count()
    # select conducting layers
    for face_id in faces_with_ports:
        connector_region = ltn.make_layer(layout.layer(default_layers[f"{face_id}_ports"]), f"connector_{face_id}")
        ltn.connect(connector_region)
    # extract netlist for the cell
    ltn.extract_netlist()
    # extract cell to circuit map for finding the netlist of interest
    cm = ltn.const_cell_mapping_into(layout, cell)
    reverse_cell_map = {v: k for k, v in cm.table().items()}
    # export the circuit of interest
    circuit = ltn.netlist().circuit_by_cell_index(reverse_cell_map[cell.cell_index()])
    if circuit:
        log.info(f"Exporting netlist to {filename}")
        _export_netlist(circuit, filename, ltn.internal_layout(), layout, cm, pcell, breakdown_list)
    else:
        log.info(f"No circuit found for {cell.display_title()}")


def _export_netlist(circuit, filename, internal_layout, original_layout, cell_mapping, pcell, breakdown_list):
    """A helper function of ``export_cell_netlist``,  exports ``circuit`` into ``filename``.

    Args:
        circuit: pya Circuit object
        filename: absolute path as convertible to string
        internal_layout: pya layout object where the netlist cells are registered
        original_layout: pya Layout object where the original cells and pcells are registered
        cell_mapping: CellMapping object as given by pya LayoutToNetlist object
        pcell: pya PCell object from which circuit was extracted
        breakdown_list: a list of Elements to break down for the netlist
    """

    # first flatten subcircuits mentioned in elements to breakdown
    # TODO implement an efficient depth first search or similar solution
    for _ in range(internal_layout.top_cell().hierarchy_levels()):
        subcircuits = list(circuit.each_subcircuit())
        for subcircuit in subcircuits:
            internal_cell = internal_layout.cell(subcircuit.circuit_ref().cell_index)
            if internal_cell.name.split('$')[0].replace('*', ' ') in breakdown_list:
                circuit.flatten_subcircuit(subcircuit)

    nets_for_export = {}
    for net in circuit.each_net():
        nets_for_export[net.expanded_name()] = extract_nets(net)

    subcircuits_for_export = {}
    used_internal_cells = set()

    # selects last cell in the layout, which will contain all instances of all cells with user properties
    *_, last_cell = original_layout.each_cell()

    # retrieve all instances in last_cell hierarchy
    # instances in original layout are identified by cell index and transformation
    # the concatenated transformation of instance's predecessors is stored as tuple's second element
    original_instances = []
    instance_queue = list(last_cell.each_inst())
    instance_queue = [(instance, pya.DCplxTrans.R0) for instance in instance_queue]
    while len(instance_queue) > 0:
        instance, instance_trans = instance_queue.pop(0)
        original_instances.append((instance, instance_trans))
        for child in instance.cell.each_inst():
            instance_queue.append((child, instance_trans * instance.dcplx_trans))

    # Indexing as defined in default_layers is not consistent with layer indexing in original_layout
    base_metal_gap_wo_grid_layer_idx_array = [idx for idx, li in
        enumerate(original_layout.layer_infos()) if li.name.endswith('_base_metal_gap_wo_grid')]

    for subcircuit in circuit.each_subcircuit():
        internal_cell = internal_layout.cell(subcircuit.circuit_ref().cell_index)
        if cell_mapping.has_mapping(internal_cell.cell_index()):
            original_cell_index = cell_mapping.cell_mapping(internal_cell.cell_index())
            possible_instances = [(i,i_trans) for i,i_trans in original_instances
                                                if i.cell.cell_index() == original_cell_index]
        else:
            log.info(('%s element has no cell mapping in %s between circuit layout and orignal layout,'
                    ' using subcircuit center point as subcircuit_location instead'), internal_cell.name, circuit.name)
            possible_instances = []

        used_internal_cells.add(internal_cell)

        if hasattr(subcircuit, "trans"):
            subcircuit_trans = subcircuit.trans
            subcircuit_location = (subcircuit.trans * subcircuit.circuit_ref().boundary).bbox().center()
        else:  # sane defaults for klayout 0.26 as it does not have `subcircuit.trans`
            subcircuit_trans = pya.DCplxTrans.R0
            subcircuit_location = pya.DPoint(0.0, 0.0)

        instances_with_eq_trans = [(i, i_trans) for i, i_trans in possible_instances
                                                if i_trans * i.dcplx_trans == subcircuit_trans]
        property_dict = {}
        correct_instance = None
        if instances_with_eq_trans:
            # Find property_dict if available
            instances_with_property_dict = [(i, i_trans) for i, i_trans in instances_with_eq_trans if i.has_prop_id()]
            if instances_with_property_dict:
                correct_instance, correct_instance_trans = instances_with_property_dict[0]
                property_dict = {key: value for (key, value) in
                    original_layout.properties(correct_instance.prop_id) if key != "id"}
            else:
                correct_instance, correct_instance_trans = instances_with_eq_trans[0]
            # Collect bounding boxes for all *_base_metal_gap_wo_grid layers
            # then construct a bigger bounding box that envelops all of them
            bboxes = []
            for idx in base_metal_gap_wo_grid_layer_idx_array:
                bbox = correct_instance.dbbox_per_layer(idx)
                if not bbox.empty():
                    bboxes.append(bbox)
            if len(bboxes) > 0:
                combined_bbox = pya.DBox(min([bbox.p1.x for bbox in bboxes]), min([bbox.p1.y for bbox in bboxes]),
                                         max([bbox.p2.x for bbox in bboxes]), max([bbox.p2.y for bbox in bboxes]))
                # subcircuit_location is the center of geometry of all *_base_metal_gap_wo_grid layers in the cell
                # we also transform the point by instance's predecessors' transformation
                subcircuit_location = correct_instance_trans * combined_bbox.center()
            else:
                log.info(('%s element has no bounding boxes in *_base_metal_gap_wo_grid layers in %s,'
                    ' using subcircuit center point as subcircuit_location instead'),
                    internal_cell.name, circuit.name)
        elif possible_instances:
            log.info(('Could not find a matching element for %s subcircuit in the orignal layout of %s,'
                    ' using subcircuit center point as subcircuit_location instead'), internal_cell.name, circuit.name)

        subcircuits_for_export[subcircuit.id()] = {
            "cell_name": internal_cell.name,
            "instance_name": correct_instance.property('id') if correct_instance else None,
            "subcircuit_origin": subcircuit_trans.disp,
            "subcircuit_location": subcircuit_location,
            "properties": property_dict,
        }

    circuits_for_export = {}
    for internal_cell in sorted(used_internal_cells, key=lambda cell: cell.name):
        circuits_for_export[internal_cell.name] = extract_circuits(cell_mapping, internal_cell, original_layout)

    chip_for_export = {}
    if pcell.pcell_declaration() is not None:
        chip_params = pcell.pcell_parameters_by_name()
        if {'frames_enabled', 'face_boxes', 'face_ids', 'box'} <= set(chip_params.keys()):
            for face in chip_params['frames_enabled']:
                face_box = chip_params['face_boxes'][int(face)]
                if face_box is None:
                    face_box = chip_params['box']
                face_id = chip_params['face_ids'][int(face)]
                chip_for_export[f'{face_id}_face_dimensions'] = face_box

    with open(str(filename), 'w') as fp:
        json.dump({
            "nets": nets_for_export,
            "subcircuits": subcircuits_for_export,
            "circuits": circuits_for_export,
            "chip": chip_for_export
        }, fp, cls=GeometryJsonEncoder, indent=4)


def extract_nets(net):
    """ Extract dictionary for net for JSON export """
    net_for_export = []
    for pin_ref in net.each_subcircuit_pin():
        net_for_export.append({
            "subcircuit_id": pin_ref.subcircuit().id(),
            "pin": pin_ref.pin().expanded_name()
        })

    return net_for_export


def extract_circuits(cell_mapping, internal_cell, layout):
    """ Extract dictionary for circuit for JSON export """

    circuit_has_cell = cell_mapping.has_mapping(internal_cell.cell_index())
    circuit_for_export = {"circuit_has_cell": circuit_has_cell}

    if circuit_has_cell:
        original_cell_index = cell_mapping.cell_mapping(internal_cell.cell_index())
        original_cell = layout.cell(original_cell_index)
        is_pcell = original_cell.is_pcell_variant()
        circuit_for_export["is_pcell"] = is_pcell
        pcell_parameters = original_cell.pcell_parameters_by_name()

        # remove regardless if exists since it does not actually include more than base anyway
        pcell_parameters.pop('refpoints', None)
        if is_pcell:
            circuit_for_export = {
                **circuit_for_export,
                **pcell_parameters
            }
        circuit_for_export["waveguide_length"] = get_cell_path_length(original_cell)
    return circuit_for_export



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math

from kqcircuits.chips.chip import Chip
from kqcircuits.qubits.qubit import Qubit
from kqcircuits.qubits.swissmon import Swissmon
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.elements.waveguide_coplanar_taper import WaveguideCoplanarTaper
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.pya_resolver import pya
from kqcircuits.util.coupler_lib import cap_params
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(Qubit, "junction_type", "fluxline_type")
class XMonsDirectCoupling(Chip):
    """The PCell declaration for an XMonsDirectCoupling chip."""

    qubit_spacing = Param(pdt.TypeDouble, "Qubit spacing", 10, unit="μm")
    arm_width_a = Param(pdt.TypeDouble, "Qubit 1 and 3 arm width", 24, unit="μm")
    arm_width_b = Param(pdt.TypeDouble, "Qubit 2 arm width", 24, unit="μm")
    rr_cpl_width = Param(pdt.TypeList, "RR to QB coupler width (um for each RR)", [24, 24, 24])

    def produce_readout_resonator(self, pos_start, end_y, length, name, c_kappa_l_fingers):
        # 86 ohms
        ro_a = 5
        ro_b = 20

        taper_length = 100

        # T to PL
        _, pl_cross_ref = self.insert_cell(WaveguideCoplanarSplitter,
            pya.DTrans(pos_start.x, end_y),
            inst_name=("PL{}".format(name) if name else None),
            label_trans=pya.DTrans.R90,
            **t_cross_parameters(a=self.a, b=self.b, a2=self.a, b2=self.b, length_extra_side=20, length_extra=0)
        )

        # Finger cap
        cplr_cell = self.add_element(**cap_params(4, c_kappa_l_fingers, "interdigital"))
        cplr_ref_rel = self.get_refpoints(cplr_cell, pya.DTrans.R90)
        _, cplr_ref = self.insert_cell(
            cplr_cell, pya.DTrans(pl_cross_ref["port_bottom"]-cplr_ref_rel["port_b"])*pya.DTrans.R90)

        # Taper to T
        taper_cell, taper_ref_rel = WaveguideCoplanarTaper.create_with_refpoints(
            self.layout, self.LIBRARY_NAME, pya.DTrans.R90,
            a=ro_a,
            b=ro_b,
            a2=self.a,
            b2=self.b,
            taper_length=taper_length
        )
        _, taper_ref = self.insert_cell(
            taper_cell, pya.DTrans(cplr_ref["port_a"]-taper_ref_rel["port_b"])*pya.DTrans.R90)

        # T to tail and straight
        rr_cross_cell = self.add_element(WaveguideCoplanarSplitter, **t_cross_parameters(
            length_extra_side=40,
            length_extra=0,
            a=ro_a,
            b=ro_b,
            a2=ro_a,
            b2=ro_b,
        ))
        rr_cross_ref_rel = self.get_refpoints(rr_cross_cell, pya.DTrans.R90)
        _, rr_cross_ref = self.insert_cell(
            rr_cross_cell, pya.DTrans(taper_ref["port_a"]-rr_cross_ref_rel["port_right"])*pya.DTrans.R90)

        # straight
        self.insert_cell(WaveguideComposite,
            nodes=[
                Node(rr_cross_ref["port_left"]),
                Node(pos_start, n_bridges=10),
            ],
            a=ro_a,
            b=ro_b,
        )

        # tail
        non_tail_length = (rr_cross_ref["base"]-pos_start).length()
        tail_r = 100
        tail_v = pya.DVector(0, -(length
                                  - non_tail_length  # straight down
                                  - (rr_cross_ref["base"]-rr_cross_ref["port_bottom"]).length()  # t-piece
                                  - math.pi/2*tail_r  # curve
                                  + tail_r  # vertical offset for the straight part with fixed length
                                  ))
        self.insert_cell(WaveguideComposite,
            nodes=[
                Node(rr_cross_ref["port_bottom"]),
                Node(rr_cross_ref["port_bottom_corner"]),
                Node(rr_cross_ref["port_bottom_corner"]+tail_v, n_bridges=5),
            ],
            a=ro_a,
            b=ro_b,
            r=tail_r
        )

    def produce_qubits(self):
        """A dedicated function to be used also by the corresponding simulation object.

        Assumes following attributes:

        * arm_width_a
        * arm_width_b
        * qubit_spacing
        * fluxline_type
        * rr_cpl_width
        * junction_type
        * layout
        * insert_cell
        """

        # Finnmon
        arm_length = 146
        qubit_props_common = {
            "arm_length": [arm_length] * 4,
            "island_r": 2,
            "port_width": [10, 5, 10],
        }
        arm_width_a = self.arm_width_a
        arm_width_b = self.arm_width_b
        full_gap_width = 72
        qb1_coupler_width, qb2_coupler_width, qb3_coupler_width = [float(param) for param in self.rr_cpl_width]
        finnmon_1 = self.add_element(Swissmon,
            arm_width=[arm_width_a] * 4,
            gap_width=[(full_gap_width - arm_width_a) / 2] * 4,
            cpl_gap=[110, 90, 110],
            cpl_length=[0, 134+qb1_coupler_width, 0],
            cpl_width=[60, qb1_coupler_width, 60],
            cl_offset=[150, -150],
            **qubit_props_common)
        finnmon_2 = self.add_element(Swissmon,
            arm_width=[arm_width_b] * 4,
            gap_width=[(full_gap_width - arm_width_b) / 2] * 4,
            cpl_gap=[110, 102, 110],
            cpl_length=[0, 116+qb2_coupler_width, 0],
            cpl_width=[60, qb2_coupler_width, 60],
            cl_offset=[150, 150],
            **qubit_props_common)
        finnmon_3 = self.add_element(Swissmon,
            arm_width=[arm_width_a]*4,
            gap_width=[(full_gap_width - arm_width_a) / 2] * 4,
            cpl_gap=[110, 90, 110],
            cpl_length=[0, 134+qb3_coupler_width, 0],
            cpl_width=[60, qb3_coupler_width, 60],
            cl_offset=[-150, -150],
            **qubit_props_common)

        qubit_y = 5e3
        qubit_step = ((full_gap_width-arm_width_a)/2+(full_gap_width-arm_width_b)/2)+2*arm_length
        self.insert_cell(finnmon_1,
                         trans=pya.DTrans(5e3 - qubit_step - self.qubit_spacing, qubit_y),
                         inst_name="QB1",
                         label_trans=pya.DTrans.R90,
                         rec_levels=None)
        self.insert_cell(finnmon_2,
                         trans=pya.DTrans(5e3, qubit_y),
                         inst_name="QB2",
                         label_trans=pya.DTrans.R90,
                         rec_levels=None
                         )
        self.insert_cell(finnmon_3,
                         trans=pya.DTrans(5e3 + qubit_step + self.qubit_spacing, qubit_y),
                         inst_name="QB3",
                         label_trans=pya.DTrans.R90,
                         rec_levels=None
                         )

    def build(self):

        self.produce_launchers("SMA8")
        self.produce_junction_tests(junction_type=self.junction_type)
        self.produce_qubits()

        # Readout resonators
        height_rr_feedline = 7.9e3
        # target parameters V2
        self.produce_readout_resonator(
            self.refpoints["QB1_port_cplr1"], height_rr_feedline, 2421.5 + 2447.1, name="1",
            c_kappa_l_fingers=60.8
        )
        self.produce_readout_resonator(
            self.refpoints["QB2_port_cplr1"], height_rr_feedline, 2436 + 2246.3, name="2",
            c_kappa_l_fingers=61.55
        )
        self.produce_readout_resonator(
            self.refpoints["QB3_port_cplr1"], height_rr_feedline, 2438.5 + 2248.2, name="3",
            c_kappa_l_fingers=58.1
        )

        # PL input cap
        pl_in_cap_cell = self.add_element(**cap_params(8, 55.4))  # 35 fF
        pl_in_cap_ref_rel = self.get_refpoints(pl_in_cap_cell)
        pl_in_cap_trans = pya.DTrans(
            pya.DPoint(self.refpoints["QB1_port_cplr1"].x - 300, height_rr_feedline) - pl_in_cap_ref_rel["port_b"])
        self.insert_cell(pl_in_cap_cell, pl_in_cap_trans, inst_name="IN")
        # Transmission lines
        tl_gap = 300
        # RR feedline
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["NW_port"]),
            Node((self.refpoints["NW_port"].x, height_rr_feedline)),
            Node(self.refpoints["IN_port_a"], n_bridges=3),
        ])
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["IN_port_b"]),
            Node(self.refpoints["PL1_port_left"], n_bridges=1),
        ])
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["PL1_port_right"]),
            Node(self.refpoints["PL2_port_left"], n_bridges=1),
        ])
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["PL2_port_right"]),
            Node(self.refpoints["PL3_port_left"], n_bridges=1),
        ])
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["PL3_port_right"]),
            Node((self.refpoints["NE_port"].x, height_rr_feedline), n_bridges=4),
            Node(self.refpoints["NE_port"]),
        ])
        # Qb1 driveline
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["WN_port"]),
            Node((self.refpoints["NW_port"].x - tl_gap, self.refpoints["WN_port"].y)),
            Node(self.refpoints["QB1_port_drive"], n_bridges=25)
        ], term2=self.b)
        # Qb1 fluxline
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["WS_port"]),
            Node(self.refpoints["WS_port_corner"]),
            Node(self.refpoints["QB1_port_flux_corner"], n_bridges=30),
            Node(self.refpoints["QB1_port_flux"])
        ])
        # Qb2 driveline
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["SW_port"]),
            Node((self.refpoints["SW_port"].x, self.refpoints["WS_port"].y - tl_gap)),
            Node(self.refpoints["QB2_port_drive"] + pya.DPoint(0, -3*self.r), n_bridges=30),
            Node(self.refpoints["QB2_port_drive"])
        ], term2=self.b)
        # Qb2 fluxline
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["SE_port"]),
            Node((self.refpoints["SE_port"].x, self.refpoints["ES_port"].y - tl_gap)),
            Node(self.refpoints["QB2_port_flux_corner"], n_bridges=30),
            Node(self.refpoints["QB2_port_flux"])
        ])
        # Qb3 driveline
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["EN_port"]),
            Node((self.refpoints["NE_port"].x + tl_gap, self.refpoints["EN_port"].y)),
            Node(self.refpoints["QB3_port_drive"], n_bridges=25)
        ], term2=self.b)
        # Qb3 fluxline
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["ES_port"]),
            Node(self.refpoints["ES_port_corner"]),
            Node(self.refpoints["QB3_port_flux_corner"], n_bridges=30),
            Node(self.refpoints["QB3_port_flux"])
        ])



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


"""PCell declaration classes for elements.

Elements represent all the different structures that form a quantum circuit. They are implemented as PCells, with the
base class of Element being PCellDeclarationHelper. After loading elements into KLayout PCell libraries, they can be
placed in a layout using the KLayout GUI or in code.

Elements contain some shared PCell parameters, including a list of refpoints which can be used to position them. They
also have methods that make it easy to create elements with different parameters and insert them to other elements.

Elements support a concept of faces, which is used for 3D-integrated chips to place shapes in layers belonging to a
certain chip face. For example, an element may create shapes in face 0 and face 1, and the ``face_ids`` parameter
of the element determines which actual chip faces the faces 0 and 1 refer to.
"""



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from enum import Enum, auto

from kqcircuits.pya_resolver import pya

class LabelOrigin(Enum):
    """Origin of the Text PCell. One of the four corners of the cell bounding box
    """
    BOTTOMLEFT = auto()
    BOTTOMRIGHT = auto()
    TOPLEFT = auto()
    TOPRIGHT = auto()

def produce_label(cell, label, location, origin, origin_offset, margin, layers, layer_protection, size=350,
                  mirror=False):
    """Produces a Text PCell accounting for desired relative position of the text respect to the given location
    and the spacing.

    Args:
        cell: container cell for the label PCell
        label: text of the label
        location: DPoint for the location of the text
        origin: LabelOrigin at which the text is located
        origin_offset: extra spacing from the location
        margin: margin of the ground grid avoidance layer around the text
        layers: list of layers where the label text is added
        layer_protection: layer where a box around the label text is added
        size: Character height in um, default 350

    Effect:
        Shapes added to the corresponding layers
    """

    layout = cell.layout()

    if not label:
        label = "A13"  # longest label on 6 inch wafer
        protection_only = True
    else:
        protection_only = False

    # text cell
    subcells = []
    for layer in layers:
        subcells.append(layout.create_cell("TEXT", "Basic", {
            "layer": layer,
            "text": label,
            "mag": size/350*500,
        }))

    # relative placement with margin
    relative_placement = {
        LabelOrigin.BOTTOMLEFT: pya.Vector(
            subcells[0].dbbox().p1.x - margin - origin_offset,
            subcells[0].dbbox().p1.y - margin - origin_offset),
        LabelOrigin.TOPLEFT: pya.Vector(
            subcells[0].dbbox().p1.x - margin - origin_offset,
            subcells[0].dbbox().p2.y + margin + origin_offset),
        LabelOrigin.TOPRIGHT: pya.Vector(
            subcells[0].dbbox().p2.x + margin + origin_offset,
            subcells[0].dbbox().p2.y + margin + origin_offset),
        LabelOrigin.BOTTOMRIGHT: pya.Vector(
            subcells[0].dbbox().p2.x + margin + origin_offset,
            subcells[0].dbbox().p1.y - margin - origin_offset),
    }[origin] * (-1)

    if mirror:
        trans = pya.DTrans(2, True, location.x - relative_placement.x, location.y + relative_placement.y)
    else:
        trans = pya.DTrans(location + relative_placement)

    if not protection_only:
        for subcell in subcells:
            cell.insert(pya.DCellInstArray(subcell.cell_index(), trans))

    # protection layer with margin
    protection = pya.DBox(pya.DPoint(subcells[0].dbbox().p1.x - margin, subcells[0].dbbox().p1.y - margin),
                          pya.DPoint(subcells[0].dbbox().p2.x + margin, subcells[0].dbbox().p2.y + margin))
    cell.shapes(layout.layer(layer_protection)).insert(
        trans.trans(protection))



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from typing import Callable, List, Optional, Type
from kqcircuits.elements.element import Element
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.port import InternalPort, EdgePort
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.util.refpoints import RefpointToInternalPort, RefpointToEdgePort, WaveguideToSimPort, JunctionSimPort

def _get_build_function(element_class, ignore_ports, transformation_from_center):
    def _build_for_element_class(self):
        simulation_cell = self.add_element(element_class,
            **{**self.get_parameters(), "junction_type": "Sim", "fluxline_type": "none"})

        element_trans = pya.DTrans(0, False, self.box.center())
        if transformation_from_center is not None:
            element_trans *= transformation_from_center(simulation_cell)
        _, refp = self.insert_cell(simulation_cell, element_trans, rec_levels=None)

        # Add ports
        port_i = 0
        for port in element_class.get_sim_ports(self):
            if ignore_ports is not None and port.refpoint in ignore_ports:
                continue

            if isinstance(port, RefpointToInternalPort):
                self.ports.append(InternalPort((port_i := port_i + 1), refp[port.refpoint], refp[port.ground_refpoint],
                                               port.resistance, port.reactance, port.inductance, port.capacitance,
                                               port.face, port.junction, port.signal_layer))
            elif isinstance(port, RefpointToEdgePort):
                self.ports.append(EdgePort((port_i := port_i + 1), refp[port.refpoint],
                                           port.resistance, port.reactance, port.inductance, port.capacitance,
                                           port.deembed_len, port.face, port.junction))
            elif isinstance(port, WaveguideToSimPort):
                towards = port.towards
                if port.towards is None:
                    towards = f"{port.refpoint}_corner"
                self.produce_waveguide_to_port(refp[port.refpoint], refp[towards], (port_i := port_i + 1),
                                               side=port.side, a=port.a, b=port.b,
                                               term1=port.term1, turn_radius=port.turn_radius,
                                               use_internal_ports=port.use_internal_ports,
                                               waveguide_length=port.waveguide_length, face=port.face,
                                               airbridge=port.airbridge)

            elif isinstance(port, JunctionSimPort):
                if self.separate_island_internal_ports:
                    self.ports.append(InternalPort((port_i := port_i + 1), refp[port.refpoint], face=port.face))
                    self.ports.append(InternalPort((port_i := port_i + 1), refp[port.other_refpoint], face=port.face))
                else:  # Junction between the islands
                    self.ports.append(
                        InternalPort((port_i := port_i + 1),
                                     *self.etched_line(refp[port.refpoint], refp[port.other_refpoint]),
                                     face=port.face, inductance=self.junction_inductance,
                                     capacitance=self.junction_capacitance, junction=True
                        )
                    )

    return _build_for_element_class

def get_single_element_sim_class(element_class: Element,
                                 ignore_ports: Optional[List[str]] = None,
                                 transformation_from_center: Optional[Callable[[pya.Cell], pya.DTrans]] = None) \
      -> Type[Simulation]:
    """Formulates a simulation class containing a single cell of a given Element class

    Args:
        element_class: an Element class for which a simulation class is returned
        ignore_ports: If list of strings is given, simulation ports will not be created for the given
            refpoints in the simulation class.
        transformation_from_center: If None, simulated element is placed in the middle of simulation's box.
            Otherwise should be a function that takes an element cell as argument and returns a DTrans object.
            The returned transformation is applied to the element cell
            after placing it in the middle of simulation's box.
            The function should not cause any side-effects, i.e. change the cell parameters
    """
    element_sim_class = type(f"SingleElementSimulationClassFor{element_class.__name__}", (Simulation, ), {
        "separate_island_internal_ports": Param(pdt.TypeBoolean,
                            "Add InternalPorts on both islands (if applicable). Use for capacitive simulations", False),
        "junction_inductance": Param(pdt.TypeList, "Junction inductance (if junction exists)", 11.497e-9, unit="H"),
        "junction_capacitance": Param(pdt.TypeList, "Junction capacitance (if junction exists)", 0.1e-15, unit="F"),

        "build": _get_build_function(element_class, ignore_ports, transformation_from_center)
    })
    add_parameters_from(element_class)(element_sim_class)
    return element_sim_class



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.element import Element
from kqcircuits.elements.finger_capacitor_taper import FingerCapacitorTaper


@add_parameters_from(FingerCapacitorTaper, "*", "taper_length")
class FingerCapacitorSquare(Element):
    """The PCell declaration for a square finger capacitor.

    Two ports with reference points. The arm leading to the finger has the same width as fingers. The feedline has
    the same length as the width of the ground gap around the coupler.

    .. MARKERS_FOR_PNG 20,-10,ground_padding 0,17,finger_width 0,5,finger_gap
    """

    a2 = Param(pdt.TypeDouble, "Width of center conductor on the other end", -1, unit="μm",
               docstring="Non-physical value '-1' means that the default size 'a' is used.")
    b2 = Param(pdt.TypeDouble, "Width of gap on the other end", -1, unit="μm",
               docstring="Non-physical value '-1' means that the default size 'b' is used.")
    finger_gap_end = Param(pdt.TypeDouble, "Gap between the finger and other pad", 3, unit="μm")
    ground_padding = Param(pdt.TypeDouble, "Ground plane padding", 20, unit="μm")
    fixed_length = Param(pdt.TypeDouble, "Fixed length of element, 0 for auto-length", 0, unit="μm")
    ground_gap_ratio = Param(pdt.TypeDouble, "Ground connection width per gap ratio", 0, unit="μm")

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def build(self):
        y_mid = self.finger_area_width() / 2
        y_left = self.a / 2
        y_right = (self.a if self.a2 < 0 else self.a2) / 2
        x_mid = self.finger_area_length() / 2
        x_left = x_mid + self.finger_width + (self.ground_padding if y_left > y_mid else 0.0)
        x_right = x_mid + self.finger_width + (self.ground_padding if y_right > y_mid else 0.0)
        x_end = x_mid + self.finger_width + self.ground_padding
        x_max = x_end + self.corner_r

        region_ground = self.get_ground_region()

        region_taper_right = pya.Region([pya.DPolygon([
            pya.DPoint(x_mid, y_mid),
            pya.DPoint(x_right, y_mid),
            pya.DPoint(x_right, y_right),
            pya.DPoint(x_max, y_right),
            pya.DPoint(x_max, -y_right),
            pya.DPoint(x_right, -y_right),
            pya.DPoint(x_right, -y_mid),
            pya.DPoint(x_mid, -y_mid)
        ]).to_itype(self.layout.dbu)])
        region_taper_left = pya.Region([pya.DPolygon([
            pya.DPoint(-x_mid, y_mid),
            pya.DPoint(-x_left, y_mid),
            pya.DPoint(-x_left, y_left),
            pya.DPoint(-x_max, y_left),
            pya.DPoint(-x_max, -y_left),
            pya.DPoint(-x_left, -y_left),
            pya.DPoint(-x_left, -y_mid),
            pya.DPoint(-x_mid, -y_mid)
        ]).to_itype(self.layout.dbu)])

        polys_fingers = []
        for i in range(self.finger_number):
            x = (i % 2) * self.finger_gap_end - x_mid
            y = i * (self.finger_width + self.finger_gap) - y_mid
            polys_fingers.append(pya.DPolygon([
                pya.DPoint(x + self.finger_length, y + self.finger_width),
                pya.DPoint(x + self.finger_length, y),
                pya.DPoint(x, y),
                pya.DPoint(x, y + self.finger_width)
            ]))

        region_fingers = pya.Region([
            poly.to_itype(self.layout.dbu) for poly in polys_fingers
        ])
        region_etch = region_taper_left + region_taper_right + region_fingers
        region_etch.round_corners(self.corner_r / self.layout.dbu, self.corner_r / self.layout.dbu, self.n)
        self.cut_region(region_etch, x_end, max(y_mid, y_left, y_right))
        self.add_waveguides(region_etch, x_end, y_left, y_right)

        region = region_ground - region_etch

        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(region)

        # protection
        region_protection = region_ground.size(self.margin / self.layout.dbu, self.margin / self.layout.dbu, 2).merged()
        self.add_protection(region_protection)

        # ports
        x_port = max(x_end, self.fixed_length / 2)
        self.add_port("a", pya.DPoint(-x_port, 0), pya.DVector(-1, 0))
        self.add_port("b", pya.DPoint(x_port, 0), pya.DVector(1, 0))

        # adds annotation based on refpoints calculated above

    def get_ground_region(self):
        """Returns the ground region for the finger capacitor."""
        finger_area_width = self.finger_area_width()
        y_mid = finger_area_width / 2 + self.ground_padding
        y_left = self.a / 2 + self.b
        y_right = (self.a if self.a2 < 0 else self.a2) / 2 + (self.b if self.b2 < 0 else self.b2)
        x_mid = self.finger_area_length() / 2 + self.finger_width
        x_left = x_mid + (self.ground_padding if y_left < y_mid else 0.0)
        x_right = x_mid + (self.ground_padding if y_right < y_mid else 0.0)
        x_end = x_mid + self.ground_padding
        x_max = x_end + self.corner_r

        region_ground = pya.Region([pya.DPolygon([
            pya.DPoint(-x_left, -y_mid),
            pya.DPoint(-x_left, -y_left),
            pya.DPoint(-x_max, -y_left),
            pya.DPoint(-x_max, y_left),
            pya.DPoint(-x_left, y_left),
            pya.DPoint(-x_left, y_mid),
            pya.DPoint(x_right, y_mid),
            pya.DPoint(x_right, y_right),
            pya.DPoint(x_max, y_right),
            pya.DPoint(x_max, -y_right),
            pya.DPoint(x_right, -y_right),
            pya.DPoint(x_right, -y_mid),
        ]).to_itype(self.layout.dbu)])

        if self.ground_gap_ratio > 0:
            x_conn = self.ground_gap_ratio * self.finger_gap_end / 2
            left_pts = []
            right_pts = []
            for i in range(self.finger_number):
                sign = 2 * (i % 2) - 1
                x = -sign * self.finger_length / 2
                y0 = i * (self.finger_width + self.finger_gap) - (finger_area_width + self.finger_gap) / 2
                y1 = y0 + self.finger_width + self.finger_gap
                y_conn = sign * self.ground_gap_ratio * self.finger_gap / 2
                left_pts += [pya.DPoint(x - x_conn, y0 - y_conn if i > 0 else -y_mid),
                             pya.DPoint(x - x_conn, y1 + y_conn if i + 1 < self.finger_number else y_mid)]
                right_pts += [pya.DPoint(x + x_conn, y0 + y_conn if i > 0 else -y_mid),
                              pya.DPoint(x + x_conn, y1 - y_conn if i + 1 < self.finger_number else y_mid)]
            region_ground -= pya.Region([pya.DPolygon(left_pts + right_pts[::-1]).to_itype(self.layout.dbu)])

        region_ground.round_corners(self.corner_r / self.layout.dbu, self.corner_r / self.layout.dbu, self.n)
        self.cut_region(region_ground, x_end, max(y_mid, y_left, y_right))
        self.add_waveguides(region_ground, x_end, y_left, y_right)

        return region_ground

    def finger_area_width(self):
        return self.finger_number * self.finger_width + (self.finger_number - 1) * self.finger_gap

    def finger_area_length(self):
        return self.finger_length + self.finger_gap_end

    def cut_region(self, region, x_max, y_max):
        cutter = pya.Region([pya.DPolygon([
            pya.DPoint(x_max, -y_max),
            pya.DPoint(x_max, y_max),
            pya.DPoint(-x_max, y_max),
            pya.DPoint(-x_max, -y_max)
        ]).to_itype(self.layout.dbu)])
        region &= cutter

    def add_waveguides(self, region, x_end, y_left, y_right):
        x_guide = self.fixed_length / 2
        if self.fixed_length != 0 and x_guide < x_end:
            raise ValueError(f"FingerCapacitorSquare parameters not compatible with fixed_length={self.fixed_length}")
        if x_guide > x_end:
            region += pya.Region([pya.DPolygon([
                pya.DPoint(-x_end, -y_left),
                pya.DPoint(-x_guide, -y_left),
                pya.DPoint(-x_guide, y_left),
                pya.DPoint(-x_end, y_left),
            ]).to_itype(self.layout.dbu)]) + pya.Region([pya.DPolygon([
                pya.DPoint(x_end, y_right),
                pya.DPoint(x_guide, y_right),
                pya.DPoint(x_guide, -y_right),
                pya.DPoint(x_end, -y_right),
            ]).to_itype(self.layout.dbu)])

    @classmethod
    def get_sim_ports(cls, simulation):
        return Element.left_and_right_waveguides(simulation)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

"""
    Helper module for pcell parameter schemas.

    Typical usage example::

        from kqcircuits.util.parameter_helper import Validator
        schema = {
            "n": {
                "type": pya.PCellParameterDeclaration.TypeInt,
                "description": "Number of elements",
                "default": 10
            }
        }
        validator = Validator(schema)
        if validator.validate({"n": 100}):
            # Do something with validated parameters...
"""

import logging

from kqcircuits.pya_resolver import pya


def normalize_rules(name, rules):
    """Normalizes rule fields corresponding to specified name.

    Adds missing optional rule fields with default settings.

    Returns:
        Dictionary containing all rule fields.
    """
    return {
        "name": name,
        "type": rules["type"],
        "description": rules["description"],
        "hidden": _get_rule("hidden", False, rules),
        "readonly": _get_rule("readonly", False, rules),
        "unit": _get_rule("unit", None, rules),
        "default": _get_rule("default", None, rules),
        "choices": _get_rule("choices", None, rules),
        "required": _get_rule("required", False, rules)
    }


class Validator():
    """Validates KLayout parameters according to specified schema.

    Attributes:
        schema: Dictionary containing rules for validating KLayout parameters.
    """

    def __init__(self, schema):
        self.schema = schema
        self.__validate_schema()

    def validate(self, parameters):
        """Validates KLayout parameters.

        Args:
            parameters: KLayout parameters.

        Returns:
            True if validation is successful.

        Raises:
            MissingParameterException: Failed to validate because a required parameter is missing.
            InvalidParameterException: Failed to validate parameter because of invalid value.
        """
        for name, rules in self.schema.items():
            rules = normalize_rules(name, rules)
            logging.debug(f"Validating {name} using rules [{rules}].")
            if name in parameters:
                value = parameters[name]
                if value is None:
                    if rules["required"]:
                        raise ValueError(_generate_missing_parameter_message(name))
                    logging.debug(f"Validated {name}.")
                    return True
                else:
                    if _is_of_type(value, rules["type"]):
                        logging.debug(f"Validated {name}.")
                        return True
                    else:
                        raise ValueError(_generate_invalid_parameter_message(name, value))
            else:
                if rules["required"]:
                    raise ValueError(_generate_missing_parameter_message(name))
                logging.debug(f"Validated {name}.")
                return True

    def __validate_schema(self):
        """Validates schema.

        Returns:
            True if validation is successful.

        Raises:
            MissingRuleFieldException: Failed to validate schema because a required rule field is missing.
            InvalidRuleValueException: Failed to validate schema because of invalid rule value.
        """
        for name, rules in self.schema.items():
            if "type" not in rules:
                raise ValueError(_generate_missing_rule_message(name, "type"))
            if "type" in rules and not _is_valid_type(rules["type"]):
                raise ValueError(_generate_invalid_rule_message(name, "type", rules["type"]))
            if "description" not in rules:
                raise ValueError(_generate_missing_rule_message(name, "description"))
            if "description" in rules and not isinstance(rules["description"], str):
                raise ValueError(_generate_invalid_rule_message(name, "description", rules["description"]))
            if "hidden" in rules and not isinstance(rules["hidden"], bool):
                raise ValueError(_generate_invalid_rule_message(name, "hidden", rules["hidden"]))
            if "readonly" in rules and not isinstance(rules["readonly"], bool):
                raise ValueError(_generate_invalid_rule_message(name, "readonly", rules["readonly"]))
            if "unit" in rules and not isinstance(rules["unit"], str):
                raise ValueError(_generate_invalid_rule_message(name, "unit", rules["unit"]))
            if "default" in rules and not _is_of_type(rules["default"], rules["type"]):
                raise ValueError(_generate_invalid_rule_message(name, "default", rules["default"]))
            if "choices" in rules and not _is_valid_choices(rules["choices"], rules["type"]):
                raise ValueError(_generate_invalid_rule_message(name, "choices", rules["choices"]))
            if "required" in rules and not isinstance(rules["required"], bool):
                raise ValueError(_generate_invalid_rule_message(name, "required", rules["required"]))
            return True


# ********************************************************************************
# PRIVATE METHODS
# ********************************************************************************


def _get_rule(name, default, rules):
    """Get rule from dictionary without raising exception.

    Args:
        name: Rule name.
        default: Default rule value.
        rules: Rules dictionary.

    Returns:
        Dictionary containing rule.
    """
    if name in rules:
        return rules[name]
    else:
        return default


def _is_valid_type(value):
    """Check if value is a KLayout parameter type.

    Args:
        value: Type value.

    Returns:
        True if value is of a Klayout parameter type.
    """
    return value is pya.PCellParameterDeclaration.TypeBoolean \
        or value is pya.PCellParameterDeclaration.TypeDouble \
        or value is pya.PCellParameterDeclaration.TypeInt \
        or value is pya.PCellParameterDeclaration.TypeLayer \
        or value is pya.PCellParameterDeclaration.TypeList \
        or value is pya.PCellParameterDeclaration.TypeNone \
        or value is pya.PCellParameterDeclaration.TypeShape \
        or value is pya.PCellParameterDeclaration.TypeString


def _is_of_type(value, rule_type):
    """Check if value is a KLayout parameter type.

    Args:
        value: Rule value.
        rule_type: Rule type.

    Returns:
        True if value is compatible with specified Klayout parameter type.
    """
    if rule_type is pya.PCellParameterDeclaration.TypeBoolean and isinstance(value, bool):
        return True
    if rule_type is pya.PCellParameterDeclaration.TypeDouble and _is_float(value):
        return True
    if rule_type is pya.PCellParameterDeclaration.TypeInt and isinstance(value, int):
        return True
    if rule_type is pya.PCellParameterDeclaration.TypeLayer and isinstance(value, pya.LayerInfo):
        return True
    if rule_type is pya.PCellParameterDeclaration.TypeList and isinstance(value, list):
        return True
    if rule_type is pya.PCellParameterDeclaration.TypeNone and value is None:
        return True
    if rule_type is pya.PCellParameterDeclaration.TypeShape and (
            isinstance(value, (pya.DBox, pya.DEdge, pya.DPoint, pya.DPolygon, pya.DPath))
    ):
        return True
    if rule_type is pya.PCellParameterDeclaration.TypeString and isinstance(value, str):
        return True

    return False


def _is_valid_choices(value, choice_type):
    """Validate choices.

    Args:
        value: Choices.
        choice_type: Choice type.

    Returns:
        True if valid choices.
    """
    return isinstance(value, list) and value and _is_valid_choice(value, choice_type)


def _is_valid_choice(value, choice_type):
    """Validate choice type.

    Args:
        value: Choice value.
        choice_type: Choice type.

    Returns:
        True if value is valid choice type.
    """
    return (isinstance(value, list)
            and value
            and len(value) == 2
            and value[0] is not None
            and value[1] is not None
            and isinstance(value[0], str)
            and isinstance(value[1], choice_type))


def _is_float(value):
    """Check if valid float.

    Args:
        value: Number value.

    Returns:
        True if value is valid float.
    """
    return (isinstance(value, float)
            or (isinstance(value, int) and value == int(float(value))))


def _generate_missing_rule_message(rule_name, key):
    """Generate missing rule error message.

    Returns:
        Message string.
    """
    return "Missing required key [{}] in schema rule [{}].".format(key, rule_name)


def _generate_invalid_rule_message(rule_name, key, value):
    """Generate missing rule error message.

    Returns:
        Message string.
    """
    return "Invalid value [{}] for [{}] in schema rule [{}].".format(value, key, rule_name)


def _generate_missing_parameter_message(parameter_name):
    """Generate missing rule error message.

    Returns:
        Message string.
    """
    return "Missing required value for parameter [{}].".format(parameter_name)


def _generate_invalid_parameter_message(parameter_name, value):
    """Generate missing rule error message.

    Returns:
        Message string.
    """
    return "Invalid value [{}] specified for parameter [{}].".format(value, parameter_name)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

"""Default layer configuration file for KQCircuits.

Defines default values related to layers and faces.

Layers have a name, ID and data type. For example ``"annotations": (220, 0)`` The data type indicates which chip they
belong to with 0 being reserved for layer that do not belong to a face or a chip. The IDs run from 0-127 for the bottom
of the chips and 128-255 for the top.

KQCircuit layers are grouped by faces and functionality. The face-numbering system works as follows.
Each face-id consists of a number, the letter "b" or "t", and another number, for example "2b1". In
a multi-chip stack, we can have more than one chip, bonded on top of each other, and the first
number denotes which of these chips it refers to. The letter "b" or "t" means that the layer is
located at either the "bottom" or "top" of that chip (in the final place after full manufacturing
process). The last number is an additional index that can be used if necessary, in case multiple
deposition layers and etching processes are needed.

The data type 0 is reserved for auxiliary layers. Annotations, refpoints and text fields are between 220-229.
Simulations use the data type 0 by creating custom layers starting from layer number 1000.

In Klayout GUI these layers are organised in view groups according to faces. Simulation and text
layer views are hidden by default. See https://www.klayout.de/doc-qt5/manual/layer_source.html

``default_layers`` is a flat dictionary mapping layer names to corresponding pya.LayerInfo objects.
While ``default_faces`` is a dictionary mapping '1t1', '2b1', '2t1' etc. to an other dictionary, a subset of
``default_layers`` in the given face. With a minor twist: in these "face-dictionaries" the keys do not
start with the face id, for example "1t1_base_metal_gap" becomes "base_metal_gap".

Layer names should not start with ``-``, this is reserved for marking exported layers to be inverted
in ``mask_export_layers``.
"""

from pathlib import Path

from kqcircuits.pya_resolver import pya
from kqcircuits.layer_cluster import LayerCluster

# These layers are present in all faces. Layer id will have +128 for top face and data type +1 between chips
_common_layers = {
    # Metal etching layers for front face (facing towards bottom chip)
    "base_metal_gap": (1, 1),

    # merged etching layer with grid  (layers 11 & 13) that defines microscopic structures such as waveguides, launchers
    "base_metal_gap_wo_grid": (2, 1),  # etching layer without grid
    "base_metal_addition": (3, 1),  # Features subtracted from layer 11 , only used during the design.
    "ground_grid": (4, 1),  # A subset of structures combined into layer 10 (e.g. ground plane perforations)

    # Occupy the area where there should be no grids, only used during the design.
    "ground_grid_avoidance": (5, 1),
}

# common layers in b and t
_common_b_t_layers = {
    **_common_layers,

    "base_metal_gap_for_EBL": (6, 1),  # Features of layer 41 that are needed for EBL
    "waveguide_path": (7, 1),  # Waveguide's metal part, used with waveguide length and DRC calculations

    # Junction layer
    "SIS_junction": (8, 1),  # Josephson junction evaporation opening
    "SIS_shadow": (9, 1),  # Josephson junction resist undercut
    "SIS_junction_2": (11, 1),

    # Airbridge layers -- potentially obsolete
    "airbridge_pads": (18, 1),  #
    "airbridge_flyover": (19, 1),  #

    "chip_dicing": (30, 1),

    # 3D integration layers
    "underbump_metallization": (20, 1),  # flip-chip bonding
    "indium_bump": (21, 1),  # flip-chip bonding
    "through_silicon_via": (22, 1),  # TSV
    "through_silicon_via_avoidance": (25, 1),  # TSV

    # Netlist
    "ports": (26, 1),  # Considered conductive in the netlist extraction
}

def _shift_layers(layers, shift_ID, shift_data_type):
    """Add a number to replicate a group of layers on a different face.

    This is a helper function so we don't have to copy-paste similar groups of layers to several
    faces. It returns a new layer group where every layer id is increased by the number ``shift``.
    """
    return {n: (v[0] + shift_ID, v[1] + shift_data_type) for n, v in layers.items()}

# organizer layers into faces
_face_layers = {}   # layer descriptions per every chip face

_face_layers['1b1'] = {
    **_common_b_t_layers,
}

_face_layers['1t1'] = {
    **_shift_layers(_common_b_t_layers, 128, 0),
}

# Top face layers
_face_layers['2b1'] = {
    **_shift_layers(_common_b_t_layers, 0, 1),    # common layers at the "top"
}

# Ceiling face layers
_face_layers['2t1'] = {
    **_shift_layers(_common_b_t_layers, 128, 1),     # same common layers at the "ceiling"
}

# Other auxiliary layers [Layer 220-229]
_aux_layers_dict = {
    "annotations": (220, 0),
    "annotations_2": (221, 0),
    "instance_names": (222, 0),
    "mask_graphical_rep": (223, 0),
    "waveguide_length": (224, 0),  # Length only, no DRC. When Waveguide leves its layer, e.g. Airbridge.
    "refpoints": (225, 0),
}

# default_faces[face_id] contains the face dictionary for the face determined by face_id.
#
# Each face dictionary should contain:
#   - key "id" with value face_id (string)
#   - for all the available layers in that face: key "Layer_name", value pya.LayerInfo object for that layer
#
default_faces = {}
for f in ('1t1', '2b1', '1b1', '2t1'):
    default_faces[f] = {n: pya.LayerInfo(i[0], i[1], f'{f}_{n}') for n, i in _face_layers[f].items()}

# pya layer information
default_layers = {n: pya.LayerInfo(i[0], i[1], n) for n, i in _aux_layers_dict.items()}
for face, layers in default_faces.items():
    default_layers.update({f'{face}_{name}': li for name, li in layers.items()})

for f in ('1t1', '2b1', '1b1', '2t1'):
    default_faces[f]['id'] = f

default_face_id = "1t1"  # face_id of the face that is used by default in some contexts

# Layer names (without face prefix) for layers exported as individual .oas files during mask layout export.
default_mask_export_layers = [
    "base_metal_gap",
    "base_metal_gap_wo_grid",
    "airbridge_pads",
    "airbridge_flyover",
]

# Layer names (without face prefix) with mask label postfix for mask label and mask covered region creation.
default_layers_to_mask = {
    "base_metal_gap_wo_grid": "1",
    "airbridge_pads": "2",
    "airbridge_flyover": "3"
}

# Layer names (without face prefix) in `layers_to_mask` for which mask covered region is not created.
default_covered_region_excluded_layers = [
    "indium_bump",
    "through_silicon_via",
]

# Layer names (without face prefix) for layers exported as bitmap files during full mask layout export (does not
# apply to individual pixels).
mask_bitmap_export_layers = [
    # "base_metal_gap_wo_grid",
    "mask_graphical_rep",
]

# Layers to hide when exporting a bitmap with "all" layers.
all_layers_bitmap_hide_layers = [default_layers[l] for l in _aux_layers_dict] + [
    default_layers["1b1_ports"],
    default_layers["1b1_base_metal_gap"],
    default_layers["1b1_ground_grid"],
    default_layers["1b1_ground_grid_avoidance"],
    default_layers["1b1_waveguide_path"],
    default_layers["1t1_ports"],
    default_layers["1t1_base_metal_gap"],
    default_layers["1t1_ground_grid"],
    default_layers["1t1_ground_grid_avoidance"],
    default_layers["1t1_waveguide_path"],
    default_layers["2b1_ports"],
    default_layers["2b1_base_metal_gap"],
    default_layers["2b1_ground_grid"],
    default_layers["2b1_ground_grid_avoidance"],
    default_layers["2b1_waveguide_path"],
    default_layers["2b1_base_metal_gap"],
    default_layers["2b1_ground_grid"],
    default_layers["2b1_ground_grid_avoidance"],
]

# Layer clusters used for exporting only certain layers together in the same file, when exporting individual chips
# during mask layout export.
# Dictionary with items "cluster name: LayerCluster".
chip_export_layer_clusters = {
    # 1b1-face
    "SIS-1b1": LayerCluster(["1b1_SIS_junction", "1b1_SIS_shadow", "1b1_SIS_junction_2"],
                            ["1b1_base_metal_gap_for_EBL"], "1b1"),
    "airbridges-1b1": LayerCluster(["1b1_airbridge_pads", "1b1_airbridge_flyover"],
                                   ["1b1_base_metal_gap_wo_grid"], "1b1"),
    # 1t1-face
    "SIS-1t1": LayerCluster(["1t1_SIS_junction", "1t1_SIS_shadow", "1t1_SIS_junction_2"],
                            ["1t1_base_metal_gap_for_EBL"], "1t1"),
    "airbridges-1t1": LayerCluster(["1t1_airbridge_pads", "1t1_airbridge_flyover"],
                                   ["1t1_base_metal_gap_wo_grid"], "1t1"),
    # 2b1-face
    "SIS-2b1": LayerCluster(["2b1_SIS_junction", "2b1_SIS_shadow", "2b1_SIS_junction_2"],
                            ["2b1_base_metal_gap_for_EBL"], "2b1"),
    "airbridges-2b1": LayerCluster(["2b1_airbridge_pads", "2b1_airbridge_flyover"],
                                   ["2b1_base_metal_gap_wo_grid"], "2b1"),
}

# default layers to use for calculating cell path lengths with get_cell_path_length()
default_path_length_layers = [
    "1b1_waveguide_path",
    "1t1_waveguide_path",
    "2b1_waveguide_path",
    "2t1_waveguide_path",
    "waveguide_length"  # AirbridgeConnection uses this
]

# default mask parameters for each face
# dict of face_id: parameters
default_mask_parameters = {
    "1b1": {
        "wafer_rad": 76200,
        "chips_map_offset": pya.DVector(-1200, 1200),
        "chip_size": 10000,
        "chip_box_offset": pya.DVector(0, 0),
        "chip_trans": pya.DTrans(),
        "dice_width": 200,
        "text_margin": 100,
        "mask_text_scale": 1.0,
        "mask_marker_offset": 50000,
    },
    "1t1": {
        "wafer_rad": 76200,
        "chips_map_offset": pya.DVector(-1200, 1200),
        "chip_size": 10000,
        "chip_box_offset": pya.DVector(0, 0),
        "chip_trans": pya.DTrans(),
        "dice_width": 200,
        "text_margin": 100,
        "mask_text_scale": 1.0,
        "mask_marker_offset": 50000,
    },
    "2b1": {
        "wafer_rad": 76200,
        "chips_map_offset": pya.DVector(-2700, 2700),
        "chip_size": 7000,
        "chip_box_offset": pya.DVector(1500, 1500),
        "chip_trans": pya.DTrans(pya.DVector(10000, 0)) * pya.DTrans().M90,
        "dice_width": 140,
        "text_margin": 100,
        "mask_text_scale": 0.7,
        "mask_marker_offset": 50000,
    },
    "2t1": {
        "wafer_rad": 76200,
        "chips_map_offset": pya.DVector(-2700, 2700),
        "chip_size": 7000,
        "chip_box_offset": pya.DVector(1500, 1500),
        "chip_trans": pya.DTrans(pya.DVector(10000, 0)) * pya.DTrans().M90,
        "dice_width": 140,
        "text_margin": 100,
        "mask_text_scale": 0.7,
        "mask_marker_offset": 50000,
    }
}

default_parameter_values = {}

default_layer_props = str(Path(__file__).resolve().parent.parent/"layer_config"/"default_layer_props.lyp")

default_chip_label_face_prefixes = {
    "1b1": "h",
    "1t1": "b",
    "2b1": "t",
    "2t1": "c",
}



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import argparse
import logging
import sys
import subprocess
from os import listdir, chdir
from os.path import isfile, join
from pathlib import Path
from kqcircuits.simulations.export.export_and_run import export_and_run
from kqcircuits.simulations.export.export_singularity import export_singularity
from kqcircuits.simulations.export.remote_export_and_run import remote_export_and_run, remote_run_only
from kqcircuits.defaults import TMP_PATH, SCRIPTS_PATH, ROOT_PATH

logging.basicConfig(level=logging.WARN, stream=sys.stdout)

def run():
    parser = argparse.ArgumentParser(description='KQC console scripts')

    subparser = parser.add_subparsers(dest="command")

    simulate_parser = subparser.add_parser('sim', help='KQC simulation \
            export and run script. Run and export with a single command!')
    mask_parser = subparser.add_parser('mask', help='Build KQC Mask.')

    singularity_parser = subparser.add_parser('singularity', help='Build and push \
                                              singularity image to remote host.')

    simulate_parser.add_argument('export_script', type=str, help='Name of the export script')
    simulate_parser.add_argument('--export-path-basename', type=str, default=None,
                                 help='The export folder will be `TMP_PATH / Path(export_path_basename)`, \
                                       if not given, then it will be `TMP_PATH / Path(args.export_script).stem`')
    simulate_parser.add_argument('-q', '--quiet', action="store_true", help='Quiet mode: No GUI')
    simulate_parser.add_argument('-e', '--export-only', action="store_true",
                                 help='Do not run simulation, only export generated files.')

    simulate_parser.add_argument('--remote', action="store_true",
                                 help='Export and run the simulations at remote host "user@host"')
    simulate_parser.add_argument('--remote-run-only', action="store_true",
                                 help='Run the simulation at remote host "user@host"')
    simulate_parser.add_argument('--kqc-remote-tmp-path', type=str, help='Path to the used tmp directory on remote')

    simulate_parser.add_argument('--detach', action="store_true",
                                 help='Detach the remote simulation from terminal, not waiting for it to finish')
    simulate_parser.add_argument('--poll-interval', type=int,
                                 help='Interval for polling the job state of remote simulation in seconds')

    mask_parser.add_argument('mask_script', type=str, help='Name of the mask script')
    mask_parser.add_argument('-d', '--debug', action="store_true", help="Debug mode. Use a single process and "
                             "print logs to standard output too.")
    mask_parser.add_argument('-m', '--mock_chips', action="store_true", help="Replace all chips with empty frames for "
                                                                             "faster testing of the mask layout")
    mask_parser.add_argument('-s', '--skip_extras', action="store_true", help="Skip netlist and documentation export")
    mask_parser.add_argument('-c N', action="store_true", help="Limit the number of used CPUs to 'N'")

    singularity_parser.add_argument('--build', action="store_true", help='build singularity image locally')
    singularity_parser.add_argument('--push', type=str, help='Destination of the export "user@host:dir"\
                                    if left empty the defaults from .remote_profile.txt will be used')
    singularity_parser.add_argument('--singularity-remote-path', type=str,
                                    help='Installation path for the singularity image on remote')

    args, args_for_script = parser.parse_known_args()

    if args.command == "sim":
        if args.export_script == "ls":
            simdir = Path(SCRIPTS_PATH / "simulations")
            for f in listdir(simdir):
                if isfile(join(simdir, f)):
                    print(f)
            return

        if args.remote:
            remote_host = str(args.export_script)
            remote_export_and_run(remote_host,
                                  args.kqc_remote_tmp_path,
                                  args.detach,
                                  args.poll_interval,
                                  args_for_script)
            return
        if args.remote_run_only:
            remote_host = str(args.export_script)
            remote_run_only(remote_host,
                            args.kqc_remote_tmp_path,
                            args.detach,
                            args.poll_interval,
                            args_for_script)
            return

        script_file = Path(args.export_script)
        if args.export_path_basename is not None:
            export_path = TMP_PATH / args.export_path_basename
        else:
            export_path = TMP_PATH / script_file.stem

        if not script_file.is_file():
            script_file = Path(SCRIPTS_PATH / "simulations" / script_file)
            if not script_file.is_file():
                logging.error(f"Export script not found at {args.export_script} or {script_file}")
                return
        export_and_run(script_file, export_path, args.quiet, args.export_only, args_for_script)
    elif args.command == "mask":
        if args.mask_script == "ls":
            maskdir = Path(SCRIPTS_PATH / "masks")
            for f in listdir(maskdir):
                if isfile(join(maskdir, f)):
                    print(f)
            return
        script_file = Path(args.mask_script)
        if not script_file.is_file():
            script_file = Path(SCRIPTS_PATH / "masks" / args.mask_script)
            if not script_file.is_file():
                logging.error(f"Mask script not found at {args.mask_script} or {script_file}")
                return
        if args.debug:
            subprocess.call([sys.executable, script_file, '-d'])
        else:  # Windows needs this for multiprocessing
            with open(script_file) as mask_file:
                exec(mask_file.read())  # pylint: disable=exec-used
    elif args.command == "singularity":
        if args.build:
            chdir(Path(ROOT_PATH / "singularity"))
            with open('install_software.sh','r') as f:
                for line in f:
                    if 'export MPI_VERSION=' in line:
                        mpi_v = line.partition('export MPI_VERSION=')[2].strip()
                        print((f"Singularity will use MPI version {mpi_v}. "
                                "Make sure this corresponds to the MPI version on the target machine\n"
                                "MPI and other package versions used by singularity can be changed "
                                "in the beginning of the /singularity/install_software.sh script"))
                        break
            subprocess.call("./singularity.sh", shell=True)
        elif args.push is not None:
            export_singularity(args.push, args.singularity_remote_path)
    else:
        parser.print_usage()



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
# pylint: disable=R0904
# TODO: Consider refactoring to reduce number of public methods

import logging
import numpy

from kqcircuits.defaults import default_layers, default_junction_type, default_sampleholders, default_mask_parameters, \
    default_bump_parameters, default_marker_type
from kqcircuits.elements.chip_frame import ChipFrame
from kqcircuits.elements.element import Element
from kqcircuits.elements.launcher import Launcher
from kqcircuits.elements.launcher_dc import LauncherDC
from kqcircuits.pya_resolver import pya
from kqcircuits.util.merge import merge_layout_layers_on_face
from kqcircuits.util.parameters import Param, pdt, add_parameters_from, add_parameter
from kqcircuits.test_structures.junction_test_pads.junction_test_pads import JunctionTestPads
from kqcircuits.test_structures.stripes_test import StripesTest
from kqcircuits.util.groundgrid import make_grid
from kqcircuits.elements.tsvs.tsv import Tsv
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_dc import FlipChipConnectorDc
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_rf import FlipChipConnectorRf


@add_parameters_from(Tsv, "tsv_type")
@add_parameters_from(FlipChipConnectorRf, "connector_type")
@add_parameter(ChipFrame, "box", hidden=True)
@add_parameters_from(ChipFrame, "name_mask", "name_chip", "name_copy", "name_brand", "chip_dicing_in_base_metal",
                     "dice_grid_margin", marker_types=[default_marker_type] * 8)
class Chip(Element):
    """Base PCell declaration for chips.

    Produces labels in pixel corners, dicing edge, markers and optionally grid for all chip faces.
    Contains methods for producing launchers in face 0 and connectors between faces 0 and 1.
    """

    LIBRARY_NAME = "Chip Library"
    LIBRARY_DESCRIPTION = "Superconducting quantum circuit library for chips."
    LIBRARY_PATH = "chips"

    with_grid = Param(pdt.TypeBoolean, "Make ground plane grid", False)
    merge_base_metal_gap = Param(pdt.TypeBoolean, "Merge grid and other gaps into base_metal_gap layer", False)
    a_capped = Param(pdt.TypeDouble, "Capped center conductor width", 10, unit="μm",
                     docstring="Width of center conductor in the capped region (μm)")
    b_capped = Param(pdt.TypeDouble, "Width of gap in the capped region ", 10, unit="μm")

    # Tsv grid parameters
    with_gnd_tsvs = Param(pdt.TypeBoolean, "Make ground TSVs", False)
    tsv_grid_spacing = Param(pdt.TypeDouble, "TSV grid distance (center to center)", 300, unit="μm")
    tsv_edge_to_tsv_edge_separation = \
        Param(pdt.TypeDouble, "Ground TSV clearance to manually placed TSVs (edge to edge)", 250, unit="μm")
    tsv_edge_to_nearest_element = Param(pdt.TypeDouble, "Ground TSV clearance to other elements (edge to edge)",
                                        100, unit="μm")
    edge_from_tsv = Param(pdt.TypeDouble, "Ground TSV center clearance to chip edge", 550, unit="μm")
    with_face1_gnd_tsvs = Param(pdt.TypeBoolean, "Make ground TSVs on the top face", False)
    with_gnd_bumps = Param(pdt.TypeBoolean, "Make ground bumps", False)
    bump_grid_spacing = Param(
        pdt.TypeDouble, "Bump grid distance (center to center)",
        default_bump_parameters['bump_grid_spacing'], unit="μm")
    bump_edge_to_bump_edge_separation = Param(
        pdt.TypeDouble, "In bump clearance to manually placed Bumps (edge to edge)",
        default_bump_parameters['bump_edge_to_bump_edge_separation'], unit="μm")
    edge_from_bump = Param(pdt.TypeDouble, "Spacing between bump and chip edge",
                           default_bump_parameters['edge_from_bump'], unit="μm")

    frames_enabled = Param(pdt.TypeList, "List of face ids (integers) for which a ChipFrame is drawn", [0])
    frames_marker_dist = Param(pdt.TypeList, "Marker distance from edge for each chip frame", [1500, 1000], unit="[μm]")
    frames_diagonal_squares = Param(pdt.TypeList, "Number of diagonal marker squares for each chip frame", [10, 2])
    frames_mirrored = Param(pdt.TypeList,
                            "List of booleans specifying if the frame is mirrored for each chip frame", [False, True])
    frames_dice_width = Param(pdt.TypeList, "Dicing street width for each chip frame", [200, 140], unit="[μm]")

    face_boxes = Param(
        pdt.TypeShape,
        "List of chip frame sizes (type DBox) for each face. None uses the chips box parameter.",
        default=[None, pya.DBox(pya.DPoint(1500, 1500), pya.DPoint(8500, 8500))],
        hidden=True)

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "{}".format(self.name_chip)

    def can_create_from_shape_impl(self):
        return self.shape.is_box()

    def parameters_from_shape_impl(self):
        self.box = pya.DBox(0, 0, self.shape.box_dwidth, self.shape.box_dheight)

    def transformation_from_shape_impl(self):
        return pya.Trans(self.shape.box_p1)

    @staticmethod
    def get_launcher_assignments(chip_cell):
        """Returns a dictionary of launcher assignments (port_id: launcher_id) for the chip.

        Args:
            chip_cell: Cell object of the chip
        """

        launcher_assignments = {}

        for inst in chip_cell.each_inst():
            port_name = inst.property("port_id")
            if port_name is not None:
                launcher_assignments[port_name] = inst.property("id")

        return launcher_assignments

    def produce_junction_tests(self, junction_type=default_junction_type):
        """Produces junction test pads in the chip.

        Args:
            junction_type: A string defining the type of junction used in the test pads.

        """
        junction_tests_w = self.add_element(JunctionTestPads,
                                            margin=50,
                                            area_height=1300,
                                            area_width=2500,
                                            junctions_horizontal=True,
                                            junction_type=junction_type,
                                            display_name="JunctionTestsHorizontal",
                                            )
        junction_tests_h = self.add_element(JunctionTestPads,
                                            margin=50,
                                            area_height=2500,
                                            area_width=1300,
                                            junctions_horizontal=True,
                                            junction_type=junction_type,
                                            display_name="JunctionTestsVertical",
                                            )
        self.insert_cell(junction_tests_h, pya.DTrans(0, False, .35e3, (10e3 - 2.5e3) / 2), "testarray_w")
        self.insert_cell(junction_tests_w, pya.DTrans(0, False, (10e3 - 2.5e3) / 2, .35e3), "testarray_s")
        self.insert_cell(junction_tests_h, pya.DTrans(0, False, 9.65e3 - 1.3e3, (10e3 - 2.5e3) / 2), "testarray_e")
        self.insert_cell(junction_tests_w, pya.DTrans(0, False, (10e3 - 2.5e3) / 2, 9.65e3 - 1.3e3), "testarray_n")

    def produce_opt_lit_tests(self):
        """Produces optical lithography test stripes at chip corners."""

        num_stripes = 20
        length = 100
        min_width = 1
        max_width = 15
        step = 3
        first_stripes_width = 2 * num_stripes * min_width

        combined_cell = self.layout.create_cell("Stripes")
        for i, width in enumerate(numpy.arange(max_width + 0.1 * step, min_width, -step)):
            stripes_cell = self.add_element(StripesTest, num_stripes=num_stripes, stripe_width=width,
                                            stripe_length=length)
            # horizontal
            combined_cell.insert(pya.DCellInstArray(stripes_cell.cell_index(),
                                                    pya.DCplxTrans(1, 0, False, -880, 2 * i * length +
                                                                   first_stripes_width - 200)))
            # vertical
            combined_cell.insert(pya.DCellInstArray(stripes_cell.cell_index(),
                                                    pya.DCplxTrans(1, 90, False,
                                                                   2 * i * length + length + first_stripes_width - 200,
                                                                   -880)))
            # diagonal
            diag_offset = 2 * num_stripes * width / numpy.sqrt(8)
            combined_cell.insert(pya.DCellInstArray(stripes_cell.cell_index(),
                                                    pya.DCplxTrans(1, -45, False, 250 + i * length - diag_offset,
                                                                   250 + i * length + diag_offset)))

        self.insert_cell(combined_cell, pya.DCplxTrans(1, 0, False, 1500, 1500))
        self.insert_cell(combined_cell, pya.DCplxTrans(1, 90, False, 8500, 1500))
        self.insert_cell(combined_cell, pya.DCplxTrans(1, 180, False, 8500, 8500))
        self.insert_cell(combined_cell, pya.DCplxTrans(1, 270, False, 1500, 8500))

    def produce_ground_grid(self):
        """Produces ground grid on all faces with ChipFrames.

        This method is called in build(). Override this method to produce a different set of chip frames.
        """
        for face in self.frames_enabled:
            self.produce_ground_on_face_grid(self.get_box(int(face)), int(face))

    def produce_ground_on_face_grid(self, box, face_id):
        """Produces ground grid in the given face of the chip.

        Args:
            box: pya.DBox within which the grid is created
            face_id (int): ID of the face where the grid is created

        """
        grid_area = box * (1 / self.layout.dbu)
        protection = pya.Region(self.cell.begin_shapes_rec(self.get_layer("ground_grid_avoidance", face_id))).merged()
        grid_mag_factor = 1
        region_ground_grid = make_grid(grid_area, protection,
                                       grid_step=10 * (1 / self.layout.dbu) * grid_mag_factor,
                                       grid_size=5 * (1 / self.layout.dbu) * grid_mag_factor)
        self.cell.shapes(self.get_layer("ground_grid", face_id)).insert(region_ground_grid)

    def produce_frame(self, frame_parameters, trans=pya.DTrans()):
        """Produces a chip frame and markers for the given face.

        Args:
            frame_parameters: PCell parameters for the chip frame
            trans: DTrans for the chip frame, default=pya.DTrans()
        """
        self.insert_cell(ChipFrame, trans, **frame_parameters)

    def merge_layout_layers_on_face(self, face, tolerance=0.004):
        """Creates "base_metal_gap" layer on given face.

        The layer shape is combination of three layers using subtract (-) and insert (+) operations:

            "base_metal_gap" = "base_metal_gap_wo_grid" - "base_metal_addition" + "ground_grid"

        Args:
            face: face dictionary containing layer names as keys and layer info objects as values
            tolerance: gap length to be ignored while merging (µm)
        """
        merge_layout_layers_on_face(self.layout, self.cell, face, tolerance)

    def merge_layout_layers(self):
        """Creates "base_metal_gap" layers on all faces.

         The layer shape is combination of three layers using subtract (-) and insert (+) operations:

            "base_metal_gap" = "base_metal_gap_wo_grid" - "base_metal_addition" + "ground_grid"

        This method is called in build(). Override this method to produce a different set of chip frames.
        """
        for i in range(len(self.face_ids)):
            self.merge_layout_layers_on_face(self.face(i))

    def produce_structures(self):
        """Produces chip frame and possibly other structures before the ground grid.

        This method is called in build(). Override this method to produce a different set of chip frames.
        """

        for i, face in enumerate(self.frames_enabled):
            face = int(face)
            frame_box = self.get_box(face)
            frame_parameters = self.pcell_params_by_name(
                ChipFrame,
                box=frame_box,
                face_ids=[self.face_ids[face]],
                use_face_prefix=len(self.frames_enabled) > 1,
                dice_width=float(self.frames_dice_width[i]),
                text_margin=default_mask_parameters[self.face_ids[face]]["text_margin"],
                marker_dist=float(self.frames_marker_dist[i]),
                diagonal_squares=int(self.frames_diagonal_squares[i]),
                marker_types=self.marker_types[i * 4:(i + 1) * 4]
            )

            if str(self.frames_mirrored[i]).lower() == 'true':  # Accept both boolean and string representation
                frame_trans = pya.DTrans(frame_box.center()) * pya.DTrans.M90 * pya.DTrans(-frame_box.center())
            else:
                frame_trans = pya.DTrans(0, 0)
            self.produce_frame(frame_parameters, frame_trans)

        if self.with_gnd_tsvs:
            self._produce_ground_tsvs(face_id=[0, 2])
        if self.with_face1_gnd_tsvs:
            tsv_box = self.get_box(1).enlarged(pya.DVector(-self.edge_from_tsv, -self.edge_from_tsv))
            self._produce_ground_tsvs(face_id=[3, 1], tsv_box=tsv_box)

        if self.with_gnd_bumps:
            self._produce_ground_bumps()

    def get_box(self, face=0):
        """
        Get the chip frame box for the specified face, correctly resolving defaults.

        Args:
            face: Integer specifying face, default 0

        Returns: pya.DBox box for the specified face
        """
        box = self.face_boxes[face]
        return box if box is not None else self.box

    def get_ground_bump_locations(self, bump_box):
        """
        Define the locations for a grid. This method returns the full grid.

        Args:
            bump_box: DBox specifying the region that should be filled with ground bumps

        Returns: list of DPoint coordinates where a ground bump can be placed
        """
        return self.make_grid_locations(bump_box, delta_x=self.bump_grid_spacing, delta_y=self.bump_grid_spacing)

    def _produce_ground_bumps(self):
        """Produces ground bumps between bottom and top face.

        The bumps avoid ground grid avoidance on both faces, and keep a minimum distance to any existing (manually
        placed) bumps.
        """
        logging.info('Starting ground bump generation')
        bump = self.add_element(FlipChipConnectorDc)

        bump_box = self.get_box(1).enlarged(pya.DVector(-self.edge_from_bump, -self.edge_from_bump))

        avoidance_layer_bottom = pya.Region(
            self.cell.begin_shapes_rec(self.get_layer("ground_grid_avoidance", 0))).merged()
        avoidance_layer_top = pya.Region(
            self.cell.begin_shapes_rec(self.get_layer("ground_grid_avoidance", 1))).merged()
        existing_bumps = pya.Region(
            self.cell.begin_shapes_rec(self.get_layer("indium_bump"))).merged()
        existing_bump_count = existing_bumps.count()
        avoidance_existing_bumps = existing_bumps.sized(self.bump_edge_to_bump_edge_separation
                                                        / self.layout.dbu)

        existing_tsvs_bottom = pya.Region(
            self.cell.begin_shapes_rec(self.get_layer("through_silicon_via", 0))).merged()
        avoidance_existing_tsvs_bottom = existing_tsvs_bottom. \
            sized((self.tsv_edge_to_nearest_element) / self.layout.dbu)
        existing_tsvs_top = pya.Region(
            self.cell.begin_shapes_rec(self.get_layer("through_silicon_via", 1))).merged()
        avoidance_existing_tsvs_top = existing_tsvs_top.sized(self.tsv_edge_to_nearest_element / self.layout.dbu)
        avoidance_region = (avoidance_layer_bottom + avoidance_layer_top + avoidance_existing_bumps +
                            avoidance_existing_tsvs_bottom + avoidance_existing_tsvs_top).merged()

        locations = self.get_ground_bump_locations(bump_box)

        # Determine the shape of the bump from its underbump metallization layer. Assumes that when merged the bump
        # contains only one polygon.
        bump_size_polygon = next(pya.Region(bump.begin_shapes_rec(self.get_layer("underbump_metallization")))
                                 .merged().each())

        # Use pya.Region logic to efficiently filter bumps which are inside the allowed region
        test_object_region = pya.Region([bump_size_polygon.moved(pya.Vector(pos.to_itype(self.layout.dbu)))
                                         for pos in locations])
        passed_object_region = test_object_region.outside(avoidance_region)
        bump_locations = [p.bbox().center().to_dtype(self.layout.dbu) for p in passed_object_region]

        for location in bump_locations:
            self.insert_cell(bump, pya.DTrans(location))

        logging.info(f'Found {existing_bump_count} existing bumps and inserted {len(bump_locations)} ground bumps, '
                        + f'totalling {existing_bump_count + len(bump_locations)} bumps.')

    def post_build(self):
        self.produce_structures()
        if self.with_grid:
            self.produce_ground_grid()
        if self.merge_base_metal_gap:
            self.merge_layout_layers()
        self._produce_instance_name_labels()

    def _produce_instance_name_labels(self):

        for inst in self.cell.each_inst():
            inst_id = inst.property("id")
            if inst_id:
                cell = self.layout.create_cell("TEXT", "Basic", {
                    "layer": default_layers["instance_names"],
                    "text": inst_id,
                    "mag": 400.0
                })
                label_trans = inst.dcplx_trans
                # prevent the label from being upside-down or mirrored
                if 90 < label_trans.angle < 270:
                    label_trans.angle += 180
                label_trans.mirror = False
                # optionally apply relative transformation to the label
                rel_label_trans_str = inst.property("label_trans")
                if rel_label_trans_str is not None:
                    rel_label_trans = pya.DCplxTrans.from_s(rel_label_trans_str)
                    label_trans = label_trans * rel_label_trans
                self.insert_cell(cell, label_trans)

    def produce_launchers(self, sampleholder_type, launcher_assignments=None, enabled=None, face_id=0):
        """Produces launchers for typical sample holders and sets chip size (``self.box``) accordingly.

        This is a wrapper around ``produce_n_launchers()`` to generate typical launcher configurations.

        Args:
            sampleholder_type: name of the sample holder type
            launcher_assignments: dictionary of (port_id: name) that assigns a name to some of the launchers
            enabled: list of enabled launchers, empty means all
            face_id: index of face_ids in which to insert the launchers

        Returns:
            launchers as a dictionary :code:`{name: (point, heading, distance from chip edge)}`

        """

        if sampleholder_type == "SMA8":  # this is special: it has default launcher assignments
            if not launcher_assignments:
                launcher_assignments = {1: "NW", 2: "NE", 3: "EN", 4: "ES", 5: "SE", 6: "SW", 7: "WS", 8: "WN"}

        if sampleholder_type in default_sampleholders:
            return self.produce_n_launchers(**default_sampleholders[sampleholder_type],
                                            launcher_assignments=launcher_assignments, enabled=enabled, face_id=face_id)
        return {}

    def produce_n_launchers(self, n, launcher_type, launcher_width, launcher_gap, launcher_indent, pad_pitch,
                            launcher_assignments=None, port_id_remap=None, launcher_frame_gap=None, enabled=None,
                            chip_box=None, face_id=0):
        """Produces n launchers at default locations and optionally changes the chip size.

        Launcher pads are equally distributed around the chip. This may be overridden by specifying
        the number of pads desired per chip side if ``n`` is an array of 4 numbers.

        Pads not in ``launcher_assignments`` are disabled by default. The ``enabled`` argument may
        override this. If neither argument is defined then all pads are enabled with default names.

        Args:
            n: number of launcher pads or an array of pad numbers per side
            launcher_type: type of the launchers, "RF" or "DC"
            launcher_width: width of the launchers
            launcher_gap: pad to ground gap of the launchers
            launcher_indent: distance between the chip edge and pad port
            pad_pitch: distance between pad centers
            launcher_frame_gap: gap of the launcher pad at the frame
            launcher_assignments: dictionary of (port_id: name) that assigns a name to some of the launchers
            port_id_remap: by default, left-most top edge launcher has port_id set to 1 and port_ids
                increment for other launchers in clockwise order.
                port_id_remap is a dictionary [1..n] -> [1..n] such that for port_id_remap[x] = y,
                x is the port_id of the launcher in default order and y is the port_id of that launcher
                in your desired order.
                For example, to flip the launcher order by chip's y-axis, set port_id_remap to
                ``{i+1: ((n - i + n/4-1) % n) + 1 for i in range(n)}``
            enabled: optional list of enabled launchers
            chip_box: optionally changes the chip size (``self.box``)
            face_id: index of face_ids in which to insert the launchers

        Returns:
            launchers as a dictionary :code:`{name: (point, heading, distance from chip edge)}`
        """

        if launcher_frame_gap is None:
            launcher_frame_gap = launcher_gap

        if chip_box is not None:
            self.box = chip_box

        if launcher_type == "DC":
            launcher_cell = self.add_element(LauncherDC, width=launcher_width, face_ids=[self.face_ids[face_id]])
        else:
            launcher_cell = self.add_element(Launcher, s=launcher_width, l=launcher_width,
                                             a_launcher=launcher_width, b_launcher=launcher_gap,
                                             launcher_frame_gap=launcher_frame_gap, face_ids=[self.face_ids[face_id]])

        pads_per_side = n
        if not isinstance(n, tuple):
            n = int((n + n % 4) / 4)
            pads_per_side = [n, n, n, n]

        dirs = (90, 0, -90, 180)
        trans = (pya.DTrans(3, 0, self.box.p1.x, self.box.p2.y),
                 pya.DTrans(2, 0, self.box.p2.x, self.box.p2.y),
                 pya.DTrans(1, 0, self.box.p2.x, self.box.p1.y),
                 pya.DTrans(0, 0, self.box.p1.x, self.box.p1.y))
        _w = self.box.p2.x - self.box.p1.x
        _h = self.box.p2.y - self.box.p1.y
        sides = [_w, _h, _w, _h]

        return self._insert_launchers(dirs, enabled, launcher_assignments, port_id_remap, launcher_cell,
                                      launcher_indent, launcher_width, pad_pitch, pads_per_side, sides, trans,
                                      face_id=face_id)

    def _insert_launchers(self, dirs, enabled, launcher_assignments, port_id_remap, launcher_cell, launcher_indent,
                          launcher_width, pad_pitch, pads_per_side, sides, trans, face_id):
        """Inserts launcher cell at predefined parameters and returns launcher cells

        """
        launcher_order_idx, launchers = 0, {}
        for np, dr, tr, si in zip(pads_per_side, dirs, trans, sides):
            for i in range(np):
                launcher_order_idx += 1
                if port_id_remap:
                    port_id = port_id_remap.get(launcher_order_idx, launcher_order_idx)
                else:
                    port_id = launcher_order_idx

                if launcher_assignments:
                    if port_id not in launcher_assignments:
                        continue
                    name = launcher_assignments[port_id]

                else:
                    name = str(port_id)

                if enabled and name not in enabled:
                    continue

                loc = tr * pya.DPoint(launcher_indent, si / 2 + pad_pitch * (i + 0.5 - np / 2))
                launchers[name] = (loc, dr, launcher_width)

                transf = pya.DCplxTrans(1, dr, False, loc)
                launcher_inst, launcher_refpoints = self.insert_cell(launcher_cell, transf, name)
                launcher_inst.set_property("port_id", port_id)
                self.add_port(name, launcher_refpoints["port"], face_id=face_id)
        return launchers

    def make_grid_locations(self, box, delta_x=100, delta_y=100):  # pylint: disable=no-self-use
        """
        Define the locations for a grid. This method returns the full grid.

        Args:
            box: DBox specifying a region for a grid
            delta_x: Int or float specifying the grid separation along the x dimension
            delta_y: Int or float specifying the grid separation along the y dimension

        Returns: list of DPoint coordinates for the grid.
        """

        # array size for bump creation
        n = int((box.p2 - box.p1).x / delta_x / 2) * 2  # force even number
        m = int((box.p2 - box.p1).y / delta_y / 2) * 2  # force even number

        locations = []
        for i in numpy.linspace(-n / 2, n / 2, n + 1):
            for j in numpy.linspace(-m / 2, m / 2, m + 1):
                locations.append(box.center() + pya.DPoint(i * delta_x, j * delta_y))
        return locations

    def get_ground_tsv_locations(self, tsv_box):
        """
        Define the locations for a grid. This method returns the full grid.

        Args:
            box: DBox specifying the region that should be filled with TSVs

        Returns: list of DPoint coordinates where a ground bump can be placed
        """
        return self.make_grid_locations(tsv_box, delta_x=self.tsv_grid_spacing, delta_y=self.tsv_grid_spacing)

    def insert_tsvs(self, tsv, translations):
        """
        Insert the tsv element at locations specified in the translations list

        Args:
            tsv: An instance of a TSV class
            translations: A list of pya.DVector elements specifying

        """
        for translation in translations:
            self.insert_cell(tsv, pya.DTrans(translation))

    def _produce_ground_tsvs(self, face_id=[0, 2], tsv_box=None): # pylint: disable=dangerous-default-value
        """Produces ground TSVs between bottom and top face.

         The TSVs avoid ground grid avoidance on both faces, and keep a minimum distance to any existing (manually
         placed) TSVs.
         """
        logging.info(f'Starting ground TSV generation on face(s) {[self.face_ids[f_id] for f_id in face_id]}')

        def region_from_layer(layer_name, f_id):
            return pya.Region(self.cell.begin_shapes_rec(self.get_layer(layer_name, f_id)))

        # Produce avoidance regions before adding tsv elements
        avoidance_region = pya.Region()
        avoidance_to_element_region = pya.Region()
        avoidance_existing_tsv_region = pya.Region()
        for f_id in face_id:
            avoidance_region += (region_from_layer("ground_grid_avoidance", f_id)
                                 + region_from_layer("through_silicon_via_avoidance", f_id))
            avoidance_to_element_region += (region_from_layer("indium_bump", f_id)
                                            + region_from_layer("base_metal_gap_wo_grid", f_id))
            avoidance_existing_tsv_region += region_from_layer("through_silicon_via", f_id)
        avoidance_region.merge()
        avoidance_to_element_region.merge()
        avoidance_existing_tsv_region.merge()
        existing_tsv_count = avoidance_existing_tsv_region.count()

        def filter_locations(filter_region, separation, input_locations):
            sized_tsv = tsv_size_polygon.sized(separation / self.layout.dbu)
            test_region = pya.Region([sized_tsv.moved(pos) for pos in input_locations])
            test_region.merged_semantics = False
            pass_region = test_region.outside(filter_region)
            output_locations = [p.bbox().center() for p in pass_region]
            return output_locations

        tsv = self.add_element(Tsv, n=self.n, face_ids=[self.face_ids[f_id] for f_id in face_id])

        # Determine the shape of the tsv from its through_silicon_via layer. Assumes that when merged the tsv
        # contains only one polygon.
        tsv_size_polygon = next(pya.Region(tsv.begin_shapes_rec(self.get_layer("through_silicon_via", face_id[0])))
                                .merged().each(), None)
        if tsv_size_polygon is None:
            tsv_size_polygon = next(pya.Region(tsv.begin_shapes_rec(self.get_layer("through_silicon_via", face_id[1])))
                                    .merged().each(), None)
            if tsv_size_polygon is None:
                raise ValueError("No TSVs found on either face")

        if tsv_box is None:
            tsv_box = self.box.enlarged(pya.DVector(-self.edge_from_tsv, -self.edge_from_tsv))

        locations = self.get_ground_tsv_locations(tsv_box)
        locations_itype = [pya.Vector(pos.to_itype(self.layout.dbu)) for pos in locations]

        locations_itype = filter_locations(avoidance_region, 0, locations_itype)
        locations_itype = filter_locations(avoidance_existing_tsv_region,
                                           self.tsv_edge_to_tsv_edge_separation, locations_itype)
        locations_itype = filter_locations(avoidance_to_element_region,
                                           self.tsv_edge_to_nearest_element, locations_itype)

        tsv_locations = [pos.to_dtype(self.layout.dbu) for pos in locations_itype]

        self.insert_tsvs(tsv, tsv_locations)

        logging.info(f'Found {existing_tsv_count} existing TSVs and inserted {len(tsv_locations)} ground TSVs, '
                        + f'totalling {existing_tsv_count + len(tsv_locations)} TSVs.')

        return tsv_locations



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import subprocess
from kqcircuits.defaults import ROOT_PATH


def export_singularity(remote_host: str, singularity_remote_path: str):

    if singularity_remote_path is None:
        singularity_remote_path = '~/KQCircuits/singularity'

    subprocess.call(['ssh', remote_host,
                    'mkdir', '-p',
                    singularity_remote_path,
                    singularity_remote_path + '/libexec',
                    singularity_remote_path + '/bin'])

    subprocess.call(['scp',
                    ROOT_PATH / 'singularity/libexec/kqclib',
                    ROOT_PATH / 'singularity/libexec/kqclib.sh',
                    remote_host + ':' + singularity_remote_path + '/libexec'])

    subprocess.call(['scp',
                     ROOT_PATH / 'singularity/create_links.sh',
                     remote_host + ':' + singularity_remote_path])


    subprocess.call(['ssh', remote_host,
                     'cd', singularity_remote_path, '&&',
                     './create_links.sh', '&&',
                     'mv', 'python', 'bin/python', '&&',
                     'rm', 'bin/paraview'])



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.element import Element
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.defaults import default_tsv_type
from kqcircuits.elements.tsvs import tsv_type_choices


class Tsv(Element):
    """Base Class for TSVs."""

    default_type = default_tsv_type

    tsv_type = Param(pdt.TypeString, "TSV type", default_tsv_type, choices=tsv_type_choices)
    tsv_diameter = Param(pdt.TypeDouble, "TSV diameter", 100, unit="μm")
    tsv_margin = Param(pdt.TypeDouble, "TSV margin", 30, unit="μm")

    @classmethod
    def create(cls, layout, library=None, tsv_type=None, **parameters):
        """Create a TSV cell in layout."""
        return cls.create_subtype(layout, library, tsv_type, **parameters)[0]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.chips.chip import Chip
from kqcircuits.test_structures.tsv_test_pattern import TsvTestPattern
from kqcircuits.elements.tsvs.tsv import Tsv
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.tsvs.tsv_ellipse import TsvEllipse


@add_parameters_from(TsvEllipse, "*", tsv_diameter=10)
class TsvTest(Chip):
    """Through silicon via test chip.

    Consists of arrays of TSVs and metrology segment for crossectional analysis.
    """

    array_layout = Param(pdt.TypeList, "Array layout for TSV in center",
                         [2, 2, 2, 6, 14, 2, 14, 14, 2, 2, 14, 14, 2, 14, 6, 2, 2, 2])
    metrology_pitch = Param(pdt.TypeDouble, "Pitch in the metrology", 50, unit="μm")
    cpw_distance = Param(pdt.TypeDouble, "CPW Placeholder distance", 100, unit="μm")
    hor_distance = Param(pdt.TypeDouble, "Horizontal pitch on TSV", 200, unit="μm")
    ver_distance = Param(pdt.TypeDouble, "Vertical pitch on TSV", 500, unit="μm")

    def build(self):
        # create cell pattern in the center
        cell_pattern = self.add_element(TsvTestPattern, tsv_array_form=self.array_layout)
        self.insert_cell(cell_pattern, pya.DCplxTrans(1, 0, False, 5000, 5000))

        # metrology cell for crossectional analysis
        min_spacing = self.tsv_diameter if self.tsv_type == "standard" else max([self.tsv_diameter,
                                                                                 self.tsv_elliptical_width])
        self.create_xsection(position=pya.DPoint(1250, 1250), array_form=[10, 10],
                             pitch=self.metrology_pitch + min_spacing)
        self.create_xsection(position=pya.DPoint(1250, 8750), array_form=[8, 8],
                             pitch=self.metrology_pitch + 2 * min_spacing)
        self.create_xsection(position=pya.DPoint(8750, 1250), array_form=[6, 6], pitch=self.metrology_pitch + 2.5 *
                                                                                       min_spacing)

    def create_xsection(self, position, array_form, pitch):
        tsv_unit = self.add_element(Tsv)

        for i in range(array_form[0]):
            for j in range(array_form[1]):
                x_position = (i - array_form[0] / 2) * pitch + position.x
                y_position = (j - array_form[0] / 2) * pitch + position.y
                self.insert_cell(tsv_unit, pya.DCplxTrans(1, 0, False, x_position, y_position))
